
package info.seasip.jsw64;

import java.io.*;


/**
 * Convert JSW32 (Dragon 32) to JSW64
 */
class DragonJSW
{
	private byte title[]     = new byte[128 + 6912];
	private byte dragonMem[] = new byte[32768];
	public int maxRoom = -1; 

	// Attributes
	public static final int AIR   = 1;
	public static final int WATER1= 5 | 8;
	public static final int WATER2= 8;
	public static final int EARTH1= 4 | 8;
	public static final int EARTH2= 4 | 16;
	public static final int FIRE  = 2;
	public static final int LRAMP = 5;
	public static final int RRAMP = 6;
	public static final int LCONV = 4;
	public static final int RCONV = 3;
	public static final int TRAMP = 6 | 16;

	Jsw64Game target = new Jsw64Game();

/* Patch to draw the detailed Dragon-style bath over the JSW48-style conveyor */
	byte bathroomPatch[] = 
	{
		0x11, (byte)0xD4, 0x78, /* A6A0 LD DE,78D4h */
		0x21, (byte)0x80, (byte)0xA6, /* A6A3 LD HL,BUF */
		0x0E, 0x04, 	   /* A6A6 LD C, 4 */
/* OUTER: */
		0x06, 0x08,	   /* A6A8 LD B,8 */
/* INNER: */
		0x7E,		   /* A6AA LD A,(HL) */
		0x12,		   /* A6AB LD (DE), A */
		0x23,		   /* A6AC INC HL */
		0x14,		   /* A6AD INC D */
		0x10, (byte)0xFA,  /* A6AE DJNZ INNER */
		0x16, 0x78,	   /* A6B0 LD D,78h */
		0x13,		   /* A6B3 A6B2 INC DE */
		0x0D,	 	   /* A6B4 DEC C */
		0x20, (byte)0xF2,  /* A6B6 DJNZ OUTER */
		0x3E, (byte)0xC9,  /* A6B8 LD A,0C9h [ret] */
		0x32, 0x1E, (byte)0x98, /* A6BA LD (981Eh),A [kill conveyor] */
		(byte)0xC9		  /* A6BD RET */
	};

	byte landingPatch[] =
	{
		0x3E, (byte)0xC3,  /* A6BE LD A,0C3h [jp] */
		0x32, 0x1E, (byte)0x98, /* A6C0 LD (981Eh),A [run conveyor] */
		(byte)0xC9		  /* A6C3 RET */
	};



	byte wobble[] = 
	{
		0x3a, (byte)0xcb, (byte)0x85,	/* LD A,(85CBh) */
		(byte)0xe6, 0x01, 		/* AND 1 */
		(byte)0xca, (byte)0x9b, (byte)0x96,	/* JP Z,969Bh */
		0x7e, 			/* LD A,(HL) */
		(byte)0xcb, 0x3f, 		/* SRL A */
		0x12, 			/* LD (DE),A */
		0x14,			/* INC D */
		0x23,			/* INC HL */
		0x10, (byte)0xf8, 		/* DJNZ WOBBLE */
		(byte)0xC9
	};

	private byte abs(byte v)
	{
		if (v >= 0) return v;
		return (byte)(-v);
	}

	private void convertVguard(byte[] src, int soffset, byte [] dest, int doffset)
	{
/* Adapted from Andrew Broad's Manic Miner -> Jet Set Willy guardian 
 * converter (which has thus gone from Java to C to Java!) */
/* [27-12-2005] Set velocity to scale the same way as JSW velocity; it 
 * looked faster in the emulator because Dragon JSW is faster overall */
		byte velocity   = src[soffset + 10];
/*		signed char velocity   = 3 * src[soffset + 10]; */
		int  top      = src[soffset + 2];
		int  bottom   = src[soffset + 3];
		int  startPos = src[soffset + 5];

	/* In the Dragon version, velocity is done separately from direction */
		if (src[soffset + 6] == 0) 
		{
			velocity = (byte)(-velocity);
		}
		if (velocity < 0)
		{
			top    += (startPos - top) % velocity;
			bottom -= (bottom   - top) % velocity;
		}
		else
		{
			bottom -= (bottom - startPos) % velocity;
			top    += (bottom - top) % velocity;
		}
		if (bottom == src[soffset + 6])
		{
			bottom -= abs(velocity);
		}
		while (startPos > bottom)
		{
			if (velocity >= 0) velocity = (byte)(-velocity);
			startPos += velocity;
		}
		if (startPos >= bottom && velocity >= 0)
		{
			velocity= (byte)(-velocity);
		}
		while (startPos < top)
		{
			if (velocity < 0) velocity = (byte)(-velocity);
			startPos += velocity;
		}
		if (startPos <= top && velocity < 0)
		{
			velocity = (byte)(-velocity);
		}
		dest[doffset + 0] = 2; // Vertical 
		dest[doffset + 1] = 8 | 6;	// Ink
		dest[doffset + 2] = (byte)((0xFF & src[soffset + 4]) / 8); // X
		dest[doffset + 3] = (byte)(startPos << 1);
		dest[doffset + 4] = (byte)(velocity << 1); 
		dest[doffset + 5] = mvsprite(src[soffset + 1]);	// Sprite
		dest[doffset + 6] = (byte)(top << 1);
		dest[doffset + 7] = (byte)(bottom << 1);
					// Animation AND mask
		dest[doffset + 1] &= 0x1F;
		switch(src[soffset + 7])
		{
			case 1:  break;
			case 2:  dest[doffset + 1] |= 0x20; break;
			case 4:  dest[doffset + 1] |= 0x60; break;
			default: dest[doffset + 1] |= 0xE0; break;
		}
		if (src[soffset + 8] != 0) 
			dest[doffset + 0] |= 0x10; // fast animation
					// Animation OR mask
		dest[doffset + 2] |= ((src[soffset + 11] & 7) << 5);
	}



	byte cellType(byte dragonCellType, int room, int [] subtype)
	{
		int elemType;
		int elemSubType;
		int idx;
		byte pattern[] = new byte[9];

		int cptr = 0x7472 + 8 * dragonCellType;
		switch(dragonCellType)
		{
			case 0x16:
			case 13:	/* Looks like fire, but isn't. 
					   Used in the MegaTree. */
			case 0: elemType = Jsw64Game.EARTH;
				elemSubType = 1;
				pattern[0] = EARTH1;
				break;
			case 1: elemType = Jsw64Game.WATER;
				elemSubType = 1;
				pattern[0] = WATER1;
			break;
			case 10:
			case 12:
			case 2: elemType = Jsw64Game.RRAMP;
				elemSubType = 1;
				pattern[0] = RRAMP;
				break;
			case 15:
			case 3: elemType = Jsw64Game.LRAMP;
				elemSubType = 1;
				pattern[0] = LRAMP;
				break;
/* The Dragon version of the Swimming Pool has two Water elements */
			case 20:
			case 16:
/* ... as does Jim's Nasium */
			case 17:
				elemType = Jsw64Game.WATER;
				elemSubType = 2;
				pattern[0] = WATER2;
				break;
/* Some Dragon rooms have trampolines. */
			case 21:
				elemType = Jsw64Game.TRAMPOLINE;
				elemSubType = 1;
				pattern[0] = TRAMP;
				break;
/* The Dragon version of the Beach has two Earth elements */
			case 8:
				elemType = Jsw64Game.EARTH;
				elemSubType = 2;
				pattern[0] = EARTH2;
				break;
			default: 
				elemType = Jsw64Game.WATER;
				elemSubType = 1;
				pattern[0] = WATER1;
				break;
		}	
		// Copy 8x8 map 
		System.arraycopy(dragonMem, cptr, pattern, 1, 8);
		target.setCellPattern(room, elemType, elemSubType, pattern);
		subtype[0] = elemSubType;
		return (byte)elemType;
	}

	byte grey[] = { 0x55, (byte)0xAA, 0x55, (byte)0xAA, 0x55, (byte)0xAA, 0x55, (byte)0xAA };
/* Conveyor used by the Dragon engine. Rows 1,2 go one way; 4,5 go the other */
	byte conv[] = { (byte)0xAA, 0x44, 0x00, (byte)0x88, 0x55, 0x00, 0x00, 0x00 };

/* Spectrum-compatible version 
	byte conv[] = { 0x44, 0x22, (byte)0x88, 0x00, 0x00, 0x00, 0x00, 0x00 };
*/

/* The following memory pages are available for sprites in JSW64. Note 
 * that this is a very inefficient way to allocate them, since in the 
 * Dragon version a sprite may not occupy an entire page. In addition,
 * I've left the Willy, Maria and Toilet pages off-limits; the Spectrum
 * versions of these pages (rather than the Dragon ones) are used. */ 
	int sprPages[] = { 0x98, 0x99, 0xA0, 0xA1, 0xA2, 0xA3, 0xA7, 0xA8, 
			    0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF, 0xB0, 
			    0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 
			    0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBF, 0xC5, 0xC6, 
			    0xC7, 0xC8, 0xC9, 0xCA, 0xCB, 0xCD, 0xCE, 0xCF, 
			    0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 
			    0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF, 
			    0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 
			    0xE8, 0xE9, 0xEA };

	void setram(int page, byte value)
	{
		target.poke(page - 0x98 + 0x8640, value);
	}

	byte mvsprite(int sprite)
	{
		int page;
		int soffset, sData;
		byte dest[] = new byte[256];
	
		if (sprite >= sprPages.length)
		{
			System.err.println("Sprite ID out of range: " 
				       + Integer.toString(sprite));
			return -1;
		}
		soffset = 0x5DAC + 2 * sprite;
		sData   = (0xFF & dragonMem[soffset + 1]) 
		        + 256 * (0xFF & dragonMem[soffset]);
		page = sprPages[sprite];
		System.arraycopy(dragonMem, sData, dest, 0, 256);
		target.storeSpritePage(page, dest);	
		return (byte)page;	
	}

	void mkroom(int nRoom)
	{
		int rBase, rPtr;
		int n, max, ct, x, y, len;
		byte rdir;
		int gCur;
		byte do_rope;
		byte pattern[] = new byte[9];
		int sct[] = new int[1];

		rBase = 0x7532 + 2 * nRoom;
		rBase = (0xFF & dragonMem[rBase]) * 256 + 
			(0xFF & dragonMem[rBase + 1]);
		rPtr = rBase;
		target.initRoom(nRoom);
	// Initialise cell patterns to defaults
		System.arraycopy(grey, 0, pattern, 1, 8);
		pattern[0] = WATER1;
		target.setCellPattern(nRoom, Jsw64Game.WATER, 1, pattern);
		pattern[0] = EARTH1;
		target.setCellPattern(nRoom, Jsw64Game.EARTH, 1, pattern);
		pattern[0] = EARTH2;
		target.setCellPattern(nRoom, Jsw64Game.EARTH, 2, pattern);
		pattern[0] = FIRE;
		target.setCellPattern(nRoom, Jsw64Game.FIRE,  1, pattern);
		pattern[0] = LRAMP;
		target.setCellPattern(nRoom, Jsw64Game.LRAMP, 1, pattern);
		pattern[0] = RRAMP;
		target.setCellPattern(nRoom, Jsw64Game.RRAMP, 1, pattern);
		pattern[0] = LCONV;
		System.arraycopy(conv, 0, pattern, 1, 8);
		target.setCellPattern(nRoom, Jsw64Game.LCONVEYOR, 1, pattern);
		pattern[0] = RCONV;
		target.setCellPattern(nRoom, Jsw64Game.RCONVEYOR, 1, pattern);
		target.setWillyColour(nRoom, 7);
		pattern[0] = WATER2;
		for (n = 1; n < 9; n++) pattern[n] = 0;
		target.setCellPattern(nRoom, Jsw64Game.WATER, 2, pattern);
		pattern[0] = AIR;
		target.setCellPattern(nRoom, Jsw64Game.AIR, 1, pattern);
		// Mimic the C converter, which set up the trampoline colours
		// but left the pattern alone.
		target.getCellPattern(nRoom, Jsw64Game.TRAMPOLINE, 1, pattern);
		pattern[0] = TRAMP;
		target.setCellPattern(nRoom, Jsw64Game.TRAMPOLINE, 1, pattern);
		if (nRoom > 74) return;	/* Dragon version has 74 rooms */

		// Do the bitmaps: Vertical strips
		max = dragonMem[rPtr++];
		for (n = 0; n < max; n++)
		{
			ct = cellType(dragonMem[rPtr], nRoom, sct);
			x  = dragonMem[rPtr + 1] - 1;
			y  = dragonMem[rPtr + 2] - 1;
			len = dragonMem[rPtr + 3];
			while (len-- != 0)
			{
				target.setCell(nRoom, x, y++, ct, sct[0]);
			}
			rPtr += 4;	
		}
//	fprintf(stderr, "Horiz: bytestream=%02x %02x %02x %02x\n",
//			dragonMem[rPtr], dragonMem[rPtr + 1], dragonMem[rPtr + 2], dragonMem[rPtr + 3]);
	// Do the bitmaps: Horizontal strips
	max = dragonMem[rPtr++];
	for (n = 0; n < max; n++)
	{
		ct = cellType(dragonMem[rPtr], nRoom, sct);
		x  = dragonMem[rPtr + 1] - 1;
		y  = dragonMem[rPtr + 2] - 1;
		len = dragonMem[rPtr + 3];
		while (len-- != 0)
		{
			target.setCell(nRoom, x++, y, ct, sct[0]);
		}
		rPtr += 4;
	}
//	fprintf(stderr, "Ramp: bytestream=%02x %02x %02x %02x\n",
//			dragonMem[rPtr], dragonMem[rPtr + 1], dragonMem[rPtr + 2], dragonMem[rPtr + 3]);
	// Do the bitmaps: Diagonal strips
	max = dragonMem[rPtr++];
	if (max != 0) 
	{
		rdir = dragonMem[rPtr++];
		for (n = 0; n < max; n++)
		{
			ct = cellType(dragonMem[rPtr], nRoom, sct);
			x  = dragonMem[rPtr + 1] - 1;
			y  = dragonMem[rPtr + 2] - 1;
			len = dragonMem[rPtr + 3];
			while (len-- != 0)
			{
				target.setCell(nRoom, x, y++, ct, sct[0]);
				if (rdir != 0) x++;
				else	  x--;
			}
			rPtr += 4;
		}
	}
	// Exits
	target.setExit(nRoom, Jsw64Game.JSW64_EXIT_U, dragonMem[rPtr++]);
	target.setExit(nRoom, Jsw64Game.JSW64_EXIT_D, dragonMem[rPtr++]);
	target.setExit(nRoom, Jsw64Game.JSW64_EXIT_L, dragonMem[rPtr++]);
	target.setExit(nRoom, Jsw64Game.JSW64_EXIT_R, dragonMem[rPtr++]);
	// Conveyor
//	fprintf(stderr, "Conveyor: bytestream=%02x %02x %02x %02x\n",
//			dragonMem[rPtr], dragonMem[rPtr + 1], dragonMem[rPtr + 2], dragonMem[rPtr + 3]);
	if (dragonMem[rPtr] > 0 && dragonMem[rPtr] < 5)
	{
		switch(dragonMem[rPtr])	/* special conveyor graphic */
		{
			case 2: pattern[0] = RCONV;
				System.arraycopy(dragonMem, 0x75F9, pattern,1,8);
				target.setCellPattern(nRoom, Jsw64Game.RCONVEYOR, 1, pattern);
				pattern[0] = LCONV;
				target.setCellPattern(nRoom, Jsw64Game.LCONVEYOR, 1, pattern);
				break;
/* (for the Bathroom, this is something of a placeholder. The true bitmap is
 * stored at A680h) */
			case 3: pattern[0] = RCONV;
				System.arraycopy(dragonMem, 0x7609, pattern,1,8);
				target.setCellPattern(nRoom, Jsw64Game.RCONVEYOR, 1, pattern);
				pattern[0] = LCONV;
				target.setCellPattern(nRoom, Jsw64Game.LCONVEYOR, 1, pattern);
				break;

		}
		if (dragonMem[rPtr + 1] == 0) ct = Jsw64Game.RCONVEYOR;	
		else		  	      ct = Jsw64Game.LCONVEYOR;
		x = dragonMem[rPtr + 2] - 1;
		y = dragonMem[rPtr + 3] - 1;
		len = dragonMem[rPtr + 4];
		/* Set up conveyor animation */
		target.setConveyor(nRoom, x, y, len, (dragonMem[rPtr + 1] != 0));
		while (len-- != 0)
		{
			target.setCell(nRoom, x++, y, ct, 1);
		}
		rPtr += 5;
	}
	else rPtr++;
	// Rope guardian flag
//	fprintf(stderr, "Rope: bytestream=%02x %02x %02x %02x\n",
//			dragonMem[rPtr], dragonMem[rPtr + 1], dragonMem[rPtr + 2], dragonMem[rPtr + 3]);
	do_rope = dragonMem[rPtr++];
//
// Leave the rope until after the other guardians have been done, because of
// the "non-adjacent ropes" bug
//
	// Name
	len = dragonMem[rPtr++];
//	fprintf(stderr, "len=%d [%-*.*s]\n", len, len, len, room);
	StringBuffer title = new StringBuffer("                                ");

	int p0 = 16 - (len / 2);
	for (n = 0; n < len; n++) 
		title.setCharAt(n + p0, (char)dragonMem[rPtr++]);
	target.setRoomTitle(nRoom, title.toString());
	// Fire
	max = dragonMem[rPtr++];
	if (max != 0)
	{
		pattern[0] = FIRE;
		System.arraycopy(dragonMem, 0x7B92 + 8 * dragonMem[rPtr],
				pattern, 1, 8);
		target.setCellPattern(nRoom, Jsw64Game.FIRE, 1, pattern);
		++rPtr;
	}
	for (n = 0; n < max; n++)
	{
		x = dragonMem[rPtr] - 1;
		y = dragonMem[rPtr + 1] - 1;
		target.setCell(nRoom, x, y, Jsw64Game.FIRE, 1);
		rPtr += 2;
	}
	// Items
	max = dragonMem[rPtr++];
	if (max != 0)
	{
		System.arraycopy(dragonMem, 0x7A8A + 8 * dragonMem[rPtr],
				pattern, 0, 8);
		target.setItemBitmap(nRoom, pattern);
		++rPtr;
	}
	for (n = 0; n < max; n++)
	{
		x = dragonMem[rPtr + 1] - 1;
		y = dragonMem[rPtr + 2] - 1;
		rPtr += 3;
	
//		fprintf(stderr, "add_item(%d, %d, %d, %d)\n", gl_room, x, y, target.poke(ITEM_COUNT]);
		target.addItem(nRoom, x, y);
	}
	max = dragonMem[rPtr++];
	byte guards[] = new byte [257]; //byte[max*8 + do_rope * 8 + 1];
	gCur = 0;
	guards[gCur] = -1;
	for (n = 0; n < max; n++)
	{
		switch(dragonMem[rPtr])
		{
//
// On Vertical Guardian Migration
//
// JSW48 checks bounds like this:
//
// 	y += speed;
// 	if (y >= lowerbound) { speed = -speed; }
// 	if (y <= upperbound) { y= upperbound; speed = -speed; }
//
// Dragon JSW uses:
//
// 	if (moving_down)
// 	{
//		if (y >= lowerbound) { moving_down = 0; y -= speed; }
//		y += speed;
// 	}
// 	else
// 	{
//		if ( y <= upperbound) { moving_down = 1; y += speed; }
//		y -= speed;
// 	}
// 	So, to get the same effect, add speed to initial Y?
//
			case 0:
				convertVguard(dragonMem, rPtr, guards, gCur);
				gCur += 8;
				break;  
			case 2:
				guards[gCur+0] = 1; // Horizontal
				guards[gCur+1] = 8 | 6;	// Ink
				guards[gCur+2] = (byte)((0xFF & dragonMem[rPtr + 4]) / 8);	// X
				guards[gCur+3] = (byte)(dragonMem[rPtr + 5] * 2);	// Y
// XXX Horizontal guards have a fixed speed. But it seems to be different 
// in Dragon and Spectrum versions.
				guards[gCur+4] = (byte)(dragonMem[rPtr + 10] * 2);	// Speed
				guards[gCur+5] = mvsprite(dragonMem[rPtr + 1]);	// Sprite
// 
// On the Dragon, bounds are specified in pixels. How best to convert to
// Spectrum characters? 
//
				guards[gCur+6] = (byte)((0xFF & (dragonMem[rPtr + 2] + dragonMem[rPtr + 10])) / 8); 
				guards[gCur+7] = (byte)((0xFF & (dragonMem[rPtr + 3] - dragonMem[rPtr + 10])) / 8);
				// Animation AND mask
				switch(dragonMem[rPtr + 7])
				{
					case 1:  break;
					case 2:  guards[gCur+1] |= 0x20; break;
					case 4:  guards[gCur+1] |= 0x60; break;
					default: guards[gCur+1] |= 0xE0; break;
				}
				// Animation OR mask
				guards[gCur+2] |= ((dragonMem[rPtr + 8] & 7) << 5);
				if (dragonMem[rPtr + 6] != 0) 
					guards[gCur+0] |= 0x80;
//	fprintf(stderr, "Room 0x%02x: Hguard ", gl_room);
//	for (x = 0; x < 12; x++) fprintf(stderr, "%02x ", dragonMem[rPtr + x]);
//	fprintf(stderr, "-> ");
//	for (x = 0; x < 8;  x++) fprintf(stderr, "%02x ", guards[gCur+x]);
//	fprintf(stderr, "\n");
				gCur += 8;
				break; 
		}
		rPtr += 12;
	}
//
// Do ropes after horizontal & vertical, but before arrows. This avoids
// the IX+9 bug in rope movement, and stops arrow-hitting-rope being taken for 
// Willy-hitting rope.
//
	if ((do_rope & 1) != 0)
	{
		// Rope guardian
		guards[gCur++] = 3;
		guards[gCur++] = 0x22;
		guards[gCur++] = 16;
		guards[gCur++] = 0;
/* Slightly shorten the rope compared to the JSW version. Otherwise it 
 * clips the scenery. */
		if (do_rope == -1) guards[gCur++] = 0x1F;
		else		   guards[gCur++] = 0x1E;
		// *gptr++ = 0x20;
		guards[gCur++] = 0;
		guards[gCur++] = (byte)0x83;
		guards[gCur++] = 0x36;
	}
	max = dragonMem[rPtr++];
	for (n = 0; n < max; n++)
	{
		if (dragonMem[rPtr + 1] == 0)
			guards[gCur++] = (byte)0x84;	// Arrow going right
		else	guards[gCur++] = (byte)0x04;	// Arrow going left
		guards[gCur++] = 6;			// Arrow colour
		guards[gCur++] = (byte)((dragonMem[rPtr]-1)* 16 + 4); // Y-coord
		guards[gCur++] = -1;			// Arrow bitmap
		if (dragonMem[rPtr + 1] == 0)	
		{
// XXX Work out if these initial X values are right
			guards[gCur++] = (byte)0xD0;	// Initial X
			guards[gCur++] = 0;
			guards[gCur++] = (byte)0x82;	// Arrow going right
		}
		else	
		{
			guards[gCur++] = 0x1C;	// Initial X
			guards[gCur++] = 0;
			guards[gCur++] = 0x41;	// Arrow going left
		}
		guards[gCur++] = 0;
		rPtr += 2;
	}
	guards[gCur++] = -1;
	// Prune array to the correct size
	byte g2[] = new byte [gCur];
	System.arraycopy(guards, 0, g2, 0, gCur);
	guards = g2;
	/* Default room entry patch: Activate conveyor animation, which the
	 * Bathroom turned off. In theory it would be sufficient to do this
	 * in the Landing only, but in practice there's WRITETYPER and so on
	 * and it has to be done everywhere */
	target.poke(0xA6BE, landingPatch);
	target.setInitPatch(nRoom, 0xA6BE);
	switch(nRoom)
	{	
		case 0: /* Bathroom - the proper shape for the bath */
			for (n = 0; n < 8; n++)
			{
				target.poke(0xA680 + n, dragonMem[0x7601 + n]);
				target.poke(0xA688 + n, dragonMem[0x7609 + n]);
				target.poke(0xA690 + n, dragonMem[0x7609 + n]);
				target.poke(0xA698 + n, dragonMem[0x7611 + n]);
			}
			target.poke(0xA6A0, bathroomPatch);
			target.setInitPatch(nRoom, 0xA6A0);
			break;

		case 4: /* Nightmare */
			target.setWillySprite(nRoom, mvsprite(0x27));
			break;
		case 33: /* Ballroom West */
			// No conveyor animation
			target.setInitPatch(nRoom, 0xA6B6);
			break;
		/* Manual guardian fixes for individual rooms. I've tried
		 * to keep these to a minimum -- only when the room is 
		 * not otherwise completable. */
		case 40: /* Under the megatree */
			guards[15] = (byte)0xAE;/* Stop the guardian 
						 * colliding with the rabbit. */
			break;
		case 57: /* Under the drive */
			guards[14] = 21;	/* Move the guardian one cell
						   to the right, so it doesn't
						   collide with the fire cell.*/
			/* (it does in Dragon32, but the Dragon engine doesn't
			 * care) */
			break;
		}
		target.putGuardians(nRoom, guards);
	}

	void setVariant(char c)
	{
		target.setVariant(c);
	}

/*
void mkrooms(int bank, unsigned base)
{
	int nr;
	FILE *fp;
	char filename[20];
	byte *dptr = roomdat;
	int maxr = 16384 / room_size();

	sprintf(filename,"rooms.j%c%d", get_format(), bank);
	for (nr = 0; nr < maxr; nr++)
	{
		mkroom(gl_room, dptr);
		dptr += room_size();
		++gl_room;
	}
	fp = fopen(filename, "wb");
	if (!fp)
	{
		perror(filename);
		return;
	}
	if (fwrite(roomhdr, 1, sizeof(roomhdr), fp) < sizeof(roomhdr) ||
	    fwrite(roomdat, 1, 16384, fp) < 16384)
	{
		perror(filename);
	}
	fclose(fp);
}
*/
	protected void loadSnap(String source)
		throws IOException, FileNotFoundException
	{
		byte [] data = new byte[65536];
		int len;
		int n;

		FileInputStream fis = new FileInputStream(source);
		len = fis.read(data);
		fis.close();

		for (n = 0; n < len; n++)
		{
			if (data[n]   == (byte)0x5E && data[n+1] == (byte)0x0C &&
			    data[n+2] == (byte)0x5E && data[n+3] == (byte)0x4C &&
			    data[n+4] == (byte)0x5E && data[n+5] == (byte)0x8C)
			{
//				System.err.println("Found JSW at offset " + Integer.toHexString(n - 0x5DAC));
				System.arraycopy(data, n - 0x5DAC,
						dragonMem, 0, 0x8000);

				return;
			}
		}
		throw new IOException("Could not find JSW sprite table in " + source);
	}


	public void save(String filename) throws IOException, 
					FileNotFoundException
	{
		target.saveTape(filename);
	}	

	public static void main(String args[])
	{
		if (args.length < 2)
		{
			System.err.println("Syntax: java DragonJSW ramdump tapfile\n");
			return;
		}
		try
		{
			DragonJSW obj = new DragonJSW();
			for (int n = 2; n < args.length; n++)
			{
				if (args[n].charAt(0) == '-' &&
				    args[n].charAt(1) == 'f')
				{
					obj.setVariant(args[n].charAt(2));
				}	
// -{number} => Select how many rooms to migrate. 
				if (args[n].charAt(0) == '-' &&
				    args[n].charAt(1) == 'n')
				{
					obj.maxRoom = Integer.parseInt(args[n].substring(2, args[n].length()));
				}
			}

			String source = args[0];
			String dest   = args[1];

			obj.load(source, "jsw.sna");
			obj.execute();
			obj.save(dest);

		}
		catch(FileNotFoundException e)
		{
			System.err.println("File not found: " + e.getMessage());
		}
		catch(IOException e)
		{
			System.err.println("I/O Exception: " + e.getMessage());
		}
		catch(NumberFormatException e)
		{
			System.err.println("-n option must be followed by a number");
		}
	}

	void load(String source, String snapshot) throws IOException, FileNotFoundException
	{
		loadSnap(source);
		target.load48(target.getVariant(), snapshot);
	}

	InputStream openPatch(String s) throws FileNotFoundException
	{
		InputStream is = getClass().getResourceAsStream(s);
		if (is == null) throw new FileNotFoundException(s);
		return is;
	}


	void execute() throws IOException, FileNotFoundException
	{
		int n;
		InputStream is;

                final byte[] elementMap =
                {
                        0, 1, 2, 3, 4, 5, 6, 7,
                        8, 9, 0, 1, 2, 2, 1, 1
                };
		target.setElementMap(elementMap);
		target.clearItems();

		if (maxRoom == -1) maxRoom = target.maxRoom();
		for (n = 0; n < maxRoom; n++) mkroom(n);
// Load the Dragon font 
		target.loadPatch(is = openPatch("dffix.bin"), 0xC100);
		is.close();
// And copy the game font over it
		for (n = 0; n < 80; n++) 
			target.poke(0xC180 + n, dragonMem[0x7930 + n]);
		for (n = 0; n < 200; n++) 
			target.poke(0xC208 + n, dragonMem[0x7980 + n]);
		for (n = 0; n < 8; n++) 
			target.poke(0xC1D0 + n, dragonMem[0x7A50 + n]);
// Make all messages uppercase 
		target.poke(0x845A, "PRESS ENTER TO START");
		target.poke(0x8554, "ITEMS COLLECTED 000 TIME 00:00 M"
				+ "GAMEOVER000 7:00A 7:00A");

// Copy the title screen scrolly 
		is = openPatch("title0.jw7");
		is.read(title);
		is.close();
		for (n = 4096; n < 5760; n++) title[n + 128] |= 0x40;
		System.arraycopy(dragonMem, 0x77AA, title, 5760, 0x174);
		target.setTitleScreen(title, 128);
	
// Memory map 
		setram(0x98, (byte)0x80);	// Patches //
		setram(0x99, (byte)0x80);	// Patches //
		setram(0x9e, (byte)0x80);	// Patches //
		setram(0x9f, (byte)0x80);	// Patches //
		setram(0xA0, (byte)0x60);	// Sprites //
		setram(0xA1, (byte)0x60);	// Sprites //
		setram(0xA2, (byte)0x60);	// Sprites //
		setram(0xA3, (byte)0x60);	// Sprites //

		for (n = 0xC0; n <= 0xC5; n++)
		{
			setram(n, (byte)0x80);	/* Patches & font */
		}
		for (n = 0xD0; n < 0xE4; n++)
		{
			setram(n, (byte)0x60);	/* Sprites */
		}
		setram(0xEB, (byte)0x80);		/* Patches */
		setram(0xEC, (byte)0x80);
/* Mono lower third */
		for (n = 0; n < 256; n++) target.poke(0x9A00 + n, (byte)0x47);
		// No colours during pause
		target.poke(0x8AE6, (byte)0);
		target.poke(0x8AE7, (byte)0);
		target.poke(0x8AE8, (byte)0);
		// Make the plinth white on the game-over screen.
		target.poke(0x8CB2, (byte)7);

//		target.poke(0xC400, (byte)9);		/* IMASPZAP */
		target.poke(0xC400, (byte)0);		/* IMAPLEB */
		target.poke(0x87EB, (byte)0);		/* Initial room */
		target.poke(0x8BA3, (byte)2);		/* First Landing */
		target.poke(0x9538, (byte)3);		/* Master Bedroom */
/* 955D: Sprite page for Maria */
		target.poke(0x957C, (byte)5);		/* Bed X */
		target.poke(0x9588, (byte)0);		/* Bathroom */
		target.poke(0x959E, (byte)0);		/* Bathroom */
		target.poke(0xC40C, (byte)58);	/* Tree root */

/* Apply Mono */
		target.poke(0x8A0B, (byte)0x3E);
		target.poke(0x8A0C, (byte)0x47);
		target.poke(0x8A0D, (byte)0x18);
		target.poke(0x8A0E, (byte)0x0B);

/* Apply Wobble at A6C4h*/
		target.poke(0x944E, (byte)0xCD);	
		target.poke(0x944F, (byte)0xC4);	
		target.poke(0x9450, (byte)0xA6);	
		target.poke(0xA6C4, wobble);
// Use substitute conveyor animation
		target.loadPatch(is = openPatch("dmm64.bin"), 0x9800);
		is.close();
		target.poke(0x89F0, (byte)0x1E);
		target.poke(0x89F1, (byte)0x98);
	}
}
