
package info.seasip.jsw64;

import java.io.*;

public class Jsw64Game
{
	// The actual game memory
	private byte runtime[]  = new byte[65536];
	private byte runtime7[] = new byte[16384];
	private byte rooms[]    = new byte[65536];
	
	/** Load a pre-built JSW64 runtime. It is assumed to be called
	 * rtime0.j*2 and live in the current directory.
	 * @deprecated This relies on the JSW64 build tree hanging around. 
	 * @param variant The variant of game engine to construct. 
	 * @exception IOException if an I/O error occurs.
	 * @exception FileNotFoundException if the file, or one of the
	 *   JSW64 resources, can't be found. */

	public void load(char variant) throws IOException,FileNotFoundException
	{
		load(variant, "");
	}

	/** Load a pre-built JSW64 runtime. It is assumed to be called
	 * rtime0.j*2.
	 * @deprecated This relies on the JSW64 build tree hanging around. 
	 * @param variant The variant of game engine to construct. 
	 * @param path The path to rtime0.j*2
	 * @exception IOException if an I/O error occurs.
	 * @exception FileNotFoundException if the file, or one of the
	 *   JSW64 resources, can't be found.
	 */
	public void load(char variant, String path) 
		throws IOException,FileNotFoundException
	{
		String name;
		FileInputStream fis;
	       
	       	name = path + "rtime0.j" + variant + "2";
		fis = new FileInputStream(name);
		fis.read(runtime, 0x7F80, 0x8080);
		fis.close();
		loadResources(variant);
		gameFormat = variant;
	}

	/** Construct an empty JSW64 game by upgrading the engine of a JSW48
	 * game.
	 * @param variant The variant of game engine to construct. */
	public void load48(char variant, String jsw48name)
		throws IOException,FileNotFoundException
	{
		gameFormat = variant;
		importJSW48(jsw48name);		
		loadResources(variant);
	}

	/** Load those bits of JSW64 not based on Software Projects code.
	 *  These include the tunes and the bank 7 code.
	 *
	 * @param variant Which variant of game we're talking about
	 * @exception IOException if an I/O error occurs.
	 * @exception FileNotFoundException if one of the resources can't
	 *   be found.
	 * 
	 **/
	private void loadResources(char variant)
		throws IOException, FileNotFoundException
	{
		loadResource("rtime7" + variant + "0.bin", runtime7, 0x1B00, 7000);
		loadResource("title.tun", runtime7, 0x352A, 0xAD6);
		loadResource("cheat.tun", runtime,  0xF902, 0xFE);
		loadResource("main.tun",  runtime,  0xF700, 0x100);
	}


	/** Load one of our resource files, stripping off a +3DOS header
	 * if there is one.
	 *
	 * @param name The resource name
	 * @param dest The byte array into which the resource should be 
	 * 		loaded.
	 * @param base The offset at which to load it
	 * @param len The number of bytes to load 
	 * @exception IOException if an I/O error occurs.
	 * @exception FileNotFoundException if the resource can't be found.
	 */ 
	private void loadResource(String filename, byte [] dest, 
		int base, int len) throws IOException, FileNotFoundException
	{
		byte [] header = new byte[128];
		int bytes;
		InputStream is;

		is = getClass().getResourceAsStream(filename);
		if (is == null) throw new FileNotFoundException(filename);
		try
		{
			Plus3dos.loadStream(is, dest, base, len);
		}
		finally
		{
			is.close();
		}
	}


	/* Save this game in +3DOS format to the current directory.
	 * Commented out because this isn't a good idea; the current
	 * directory probably contains our source files, several of 
	 * which have the same names.
	public void savePlus3dos() throws IOException
	{
		savePlus3dos("");
	}

	public void savePlus3dos(String path) throws IOException
	{
		String name;
		FileOutputStream fos;

		name = path + "rtime.j" + gameFormat + "2";
		fos = new FileOutputStream(name);
		fos.write(runtime, 0x7F80, 0x8080);
		fos.close();
		name = path + "rtime7" + gameFormat + ".bin";
		fos = new FileOutputStream(name);
		fos.write(runtime7, 0x1B00, 7000);
		fos.close();
		name = path + "rooms.j" + gameFormat + "1";
		fos = new FileOutputStream(name);
		fos.write(roomHeader);
		fos.write(rooms, 0, 0x4000);
		fos.close();
		name = path + "rooms.j" + gameFormat + "3";
		fos = new FileOutputStream(name);
		fos.write(roomHeader);
		fos.write(rooms, 0x4000, 0x4000);
		fos.close();
		name = path + "rooms.j" + gameFormat + "4";
		fos = new FileOutputStream(name);
		fos.write(roomHeader);
		fos.write(rooms, 0x8000, 0x4000);
		fos.close();
		name = path + "rooms.j" + gameFormat + "6";
		fos = new FileOutputStream(name);
		fos.write(roomHeader);
		fos.write(rooms, 0xC000, 0x4000);
		fos.close();
	}
*/

	/** Save the current game in TAP format. 
	 * @param tapefile The name of the TAP file to save.
	 * @exception IOException if an I/O error occurs.
	 * @exception FileNotFoundException if a resource necessary to build
	 *  the tapefile cannot be found. */
	public void saveTape(String tapefile) 
		throws IOException,FileNotFoundException
	{
		FileOutputStream fos;
	
		fos = new FileOutputStream(tapefile);
		TapeWriter tw = new TapeWriter(fos);

		try
		{
			writeTapeFile(tw, "jsw64" + gameFormat + ".bas");
			writeHiload(tw);

			tw.writeFile("rooms.j" + gameFormat + "6", rooms, 0xC000, 0x4000, 0xC000);
			tw.writeFile("rooms.j" + gameFormat + "4", rooms, 0x8000, 0x4000, 0xC000);
			tw.writeFile("rooms.j" + gameFormat + "3", rooms, 0x4000, 0x4000, 0xC000);
			tw.writeFile("rooms.j" + gameFormat + "1", rooms, 0x0000, 0x4000, 0xC000);
			tw.writeFile("title.j" + gameFormat + "7", runtime7, 0, 0x1B00, 0xC000, 0x8000);
			tw.writeFile("rtime.j" + gameFormat + "2", runtime, 0x8000, 0x8000, 0x0000);
			writeTapeFile(tw, "main.tun"); //, runtime, 0xF700, 0x100, 0xF700);
			tw.writeFile("cheat.tun", runtime, 0xF902, 0xFE, 0xF902);
			writeTapeFile(tw, "tune.bin");
			tw.writeFile("title.tun", runtime7, 0x352A, 0xAD6, 0x752A);
		}
		catch(Plus3dosException e)
		{
			System.err.println("This should never happen");
		}
		finally
		{
			fos.close();
		}
	}

	/* Write one of our internal resources out to a tapefile */
	private void writeTapeFile(TapeWriter tw, String filename)
		throws IOException, FileNotFoundException, Plus3dosException
	{
		InputStream is = getClass().getResourceAsStream(filename);
		if (is == null) throw new FileNotFoundException(filename);
		try
		{
			tw.writeStream(filename, is);
		}
		finally
		{
			is.close();
		}
		
	}

	/* Create HILOAD.J*7 and write that out */
	private void writeHiload(TapeWriter tw)
		throws IOException,FileNotFoundException
	{
		int n;
		byte [] data = new byte[7597];
		int len;
		String filename = "hiload.j" + getVariant() + "7";	
		InputStream is = getClass().getResourceAsStream("hiload64.bin");
		if (is == null) throw new FileNotFoundException(filename);
		try
		{
			len = is.read(data, 7000, 597);
		}
		finally
		{
			is.close();
		}
		System.arraycopy(runtime7, 0x1B00, data, 0, 7000);
		tw.writeFile(filename, data, 0, data.length, 0);
	}	

	private char gameFormat = 'z';
/*
	private byte roomHeader[] = {
	 'P',  'L',  'U',  'S',  '3',  'D',  'O',  'S', 
	0x1A, 0x01, 0x00, (byte)0x80, 0x40, 0x00, 0x00, 0x03,
	0x00, 0x40, 0x00, (byte)0xC0, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3B,
	};
*/

/* Default patch vectors:
 * => Default portal vector is to go to a new room
 * => Default others are to RET. 
 */
	private final byte no_patch[] =
	{
       		 (byte)0xA0, (byte)0x86, (byte)0x9F, 
		 (byte)0x86, (byte)0x9F, (byte)0x86
	};

/* CONSTANTS */
// JSW48 element type offsets 
	public static final int JSW48_ELEM_BASE   = 0xA0;
	public static final int JSW48_ELEM_AIR    = 0xA0;
	public static final int JSW48_ELEM_WATER  = 0xA9;
	public static final int JSW48_ELEM_EARTH  = 0xB2;
	public static final int JSW48_ELEM_FIRE   = 0xBB;
	public static final int JSW48_ELEM_RAMP   = 0xC4;
	public static final int JSW48_ELEM_CONVEY = 0xCD;
// Other offsets in a room : JSW48
	public static final int JSW48_TITLE       = 0x80; // Room title
	public static final int JSW48_RAMP_DIR    = 0xDA; // Ramp direction
	public static final int JSW48_RAMP_POS    = 0xDB; // Ramp position
	public static final int JSW48_RAMP_LEN    = 0xDD; // Ramp length
	public static final int JSW48_CONVEY_DIR  = 0xD6; // Conveyor direction
	public static final int JSW48_CONVEY_POS  = 0xD7; // Conveyor position
	public static final int JSW48_CONVEY_LEN  = 0xD9; // Conveyor length
	public static final int JSW48_BORDER	  = 0xDE; // Border/Willy/flags
	public static final int JSW48_ELEM_ITEM	  = 0xE1; // Item bitmap 
	public static final int JSW48_GUARDIANS   = 0xF0; // Guardian instances
// Other offsets in a room : JSW64
	public static final int JSW64_TITLE       = 0xB6; // Title
	public static final int JSW64_CONVEY_DIR  = 0xD6; // Conveyor direction
	public static final int JSW64_CONVEY_POS  = 0xD7; // Conveyor position
	public static final int JSW64_CONVEY_LEN  = 0xD9; // Conveyor length
	public static final int JSW64_GPTR	  = 0xDA; // Guardian table ptr
	public static final int JSW64_SOLAR_X	  = 0xDC; // Solar power start 
	public static final int JSW64_SOLAR_A	  = 0xDD; // Solar power attr
	public static final int JSW64_FLAGS	  = 0xDE; // Border/Willy/flags
	public static final int JSW64_AIR_MAJOR	  = 0xDF; // Air supply, major
	public static final int JSW64_AIR_MINOR	  = 0xE0; // Air supply, minor
	public static final int JSW64_ELEM_ITEM	  = 0xE1; // Item bitmap 
	public static final int JSW64_EXIT_L	  = 0xE9; // Left exit 
	public static final int JSW64_EXIT_R	  = 0xEA; // Right exit 
	public static final int JSW64_EXIT_U	  = 0xEB; // Up exit 
	public static final int JSW64_EXIT_D	  = 0xEC; // Down exit 
	public static final int JSW64_WILLY_SPRITE= 0xED; // Willy's sprite 
	public static final int JSW64_PTL_SPRITE  = 0xEE; // Portal sprite 
	public static final int JSW64_PTL_POS	  = 0xF0; // Portal attr pos
	public static final int	JSW64_PTL_POS2	  = 0xF2; // Portal pixel pos
	public static final int	JSW64_PTL_ATTR	  = 0xF4; // Portal attr
	public static final int JSW64_PTL_TARGET  = 0xF5; // Portal target rm
	public static final int JSW64_PTL_NEWXY	  = 0xF6; // Portal target xy
	public static final int JSW64_PTL_PATCH	  = 0xF9; // Portal ptch vector
	public static final int JSW64_SETUP_PATCH = 0xFB; // Room setup pv
	public static final int JSW64_MAIN_PATCH  = 0xFD; // Main loop pv
	public static final int JSW64_FLAGS2      = 0xFF; // More room flags

/* Manic Miner room offsets */
	public static final int MM_TITLE	= 0x200; // Room title 
	public static final int MM_CONVEY_DIR	= 0x26F; // Conveyor direction 
	public static final int MM_CONVEY_POS	= 0x270; // Conveyor position 
	public static final int MM_CONVEY_LEN	= 0x272; // Conveyor length 
	public static final int MM_BORDER	= 0x273; // Border colour 
	public static final int MM_ITEMS	= 0x275; // Item definitions 
	public static final int MM_PORTAL_ATTR	= 0x28F; // Portal attribute 
	public static final int MM_PORTAL_POS	= 0x2B0; // Portal position 
	public static final int MM_ELEM_ITEM	= 0x2B4; // Item graphic 
	public static final int MM_AIR_MAJOR	= 0x2BC; // Air remaining, maj
	public static final int MM_AIR_MINOR	= 0x2BD; // Air remaining, min
	public static final int MM_HGUARDS 	= 0x2BE; // Horizontal guards 
	public static final int MM_VGUARDS 	= 0x2DD; // Vertical guards 

/* Manic Miner element type offsets */
	public static final int MM_ELEM_BASE     = 0x220;
	public static final int MM_ELEM_AIR      = MM_ELEM_BASE;
	public static final int MM_ELEM_WATER    = MM_ELEM_BASE+9;
	public static final int MM_ELEM_CRUMBLY  = MM_ELEM_BASE+18;
	public static final int MM_ELEM_EARTH    = MM_ELEM_BASE+27;
	public static final int MM_ELEM_CONVEY   = MM_ELEM_BASE+36;
	public static final int MM_ELEM_FIRE     = MM_ELEM_BASE+45;
	public static final int MM_ELEM_FIRE2    = MM_ELEM_BASE+54;
	public static final int MM_ELEM_WATER2   = MM_ELEM_BASE+63;

/* JSW64 Cell types */
	public static final int AIR	    = 0;
	public static final int WATER       = 1;
	public static final int EARTH       = 2;
	public static final int FIRE	    = 3;
	public static final int LRAMP       = 4;
	public static final int RRAMP       = 5;
	public static final int LCONVEYOR   = 6;
	public static final int RCONVEYOR   = 7;
	public static final int CRUMBLY     = 8;
	public static final int TRAMPOLINE  = 9;
	public static final int TRAP	    = 10;

/* Manic Miner room flags */
	public static final int MM_FLAG_EUGENE = 0x80;
	public static final int MM_FLAG_SKYLAB = 0x40;
	public static final int MM_FLAG_VGUARD = 0x20;
	public static final int MM_FLAG_KONG   = 0x10;
	public static final int MM_FLAG_SOLAR  = 0x08;

/* JSW64 room flags */
	public static final int JSW64_FLAG_KAMIKAZE = 0x01;

/* Final barrier options */
	public static final int FINAL_BARRIER_TITLETOP = 1;
	public static final int FINAL_BARRIER_TITLEBOT = 2;
	public static final int FINAL_BARRIER_BUF1     = 3;
	public static final int FINAL_BARRIER_BUF2     = 4;

/* Addresses in the JSW engine */
	/** Address of initial room */
	private static final int JSW64_INITIAL_ROOM  = 0x87EB;  
	/** Address of initial Willy Y */
	private static final int JSW64_INITIAL_POS1  = 0x87E6;  
	/** Address of initial Willy X and Y */
	private static final int JSW64_INITIAL_POS2  = 0x87F0;  
	/** Address of number of items in game */
	private static final int JSW64_ITEM_COUNT    = 0x85CA;  
	/** Address of item locations table */
	private static final int JSW64_ITEM_TABLE    = 0xA400;	
	/** Address of second item table for 128-room games */
	private static final int JSW64_ITEM_TABLE2   = 0xC000;	

/* FUNCTIONS */
	// 16-bit peek and poke functions
	public static int peek_w(byte [] array, int offset)
	{
		int l = array[offset];
		int h = array[offset + 1];
		return (h << 8) | l;
	}


	public static void poke_w(byte [] array, int offset, int value)
	{
		array[offset]     = (byte)(value & 0xFF);
		array[offset + 1] = (byte)((value >> 8) & 0xFF);
	}

	/** Mirror a byte.
	 * @param b The byte to mirror
	 * @return The mirrored version of the parameter */
	public static byte mirror(byte b)
	{
		byte r = 0, sm = 1, dm = (byte)0x80;

		for (int n = 0; n < 8; n++)
		{
			if ((b & sm) != 0) r |= dm;
			dm = (byte)(dm >> 1);
			sm = (byte)(sm << 1);
		}
		return r;
	}

	/** Return the offset of the cell definitions */
	public int cellBase()
	{
		switch(gameFormat)
		{
			case 'v': case 'x': return 0x6E;
			case '[': case '$': return 0x26;
		}
		return 0x41;
	}

	/** Return the number of supported cell types */
	public int cellTypes()
	{
		switch(gameFormat)
		{
			case 'v': case 'x': return 8;
			case '[': case '$': return 16;
			default:  return 13;
		}
	}


	/** Return the number of bytes in a room definition */
	public int roomSize()
	{
		if (gameFormat == 'v' || gameFormat == 'w') return 512;
		return 1024;
	}

	/** Return the number of bits used to define a cell */
	public int roomCellBits()
	{
		switch(gameFormat)
		{
			case 'v': case 'x':	return 3;
			case 'z':		return 8;
		}
		return 4;
	}

	/** Get the variant in use */
	public char getVariant()
	{
		return gameFormat;
	}

	/** Set the variant to use. 
	 *
	 * @param f The JSW64 variant to use
	 * @exception IllegalArgumentException Thrown if the parameter is not
	 * a recognised JSW64 variant.*/
	public void setVariant(char f) throws IllegalArgumentException
	{
		switch(Character.toLowerCase(f))
		{
			case 'v': 
			case 'w': 
			case 'x':
			case 'y':
			case 'z':
			case '[':
			case '$':
			gameFormat = f;
			return;
		}
		throw (new IllegalArgumentException(f + " is not a valid JSW64 variant"));
	}

	public int roomBits()
	{
		return 0x100 + patchSpace();
	}

	public int roomBase(int nr)
	{
		switch(gameFormat)
		{
			case 'v': 
			case 'w': return nr * 0x200;
			default:  return nr * 0x400;
		}	
	}

	/** Return the maximum number of rooms in this variant. */
	public int maxRoom()
	{
		switch(gameFormat)
		{
			case 'v': 
			case 'w': return 128;
			default:  return 64;
		}
	}

	public int patchSpace()
	{
		switch(gameFormat)
		{
			case 'v': return  0x40;
			case 'w': return  0x00;
			case 'x': return 0x240;
			case 'y': return 0x200;
			case 'z': return 0x100;
			case '$':
			case '[': return 0x200;
		}
		return 0;
	}

	public void initRoom(int room)
	{
		int n, m;
		int rBase = roomBase(room);

		for (n = roomSize() - 1; n >= 0; n--) rooms[rBase + n] = 0;
		System.arraycopy(no_patch, 0, rooms, rBase + JSW64_PTL_PATCH, 
				no_patch.length);
		for (n = 0; n < 32; n++) rooms[rBase + JSW64_TITLE  + n] = 0x20;
		poke_w(rooms, rBase + JSW64_GPTR, 0x8000);
		rooms[rBase + JSW64_AIR_MAJOR] = (byte)0xFF;
		rooms[rBase + JSW64_AIR_MINOR] = (byte)0xFF;
		rooms[rBase + 0] = (byte)0xFF;	// No guardians	
		m = cellTypes();
		for (n = 0; n < m; n++)
		{
			cellSetType(room, n, n);
			rooms[rBase + cellBase() + 9 * n] = (byte)0x80;
				/* Flashing black on black */
		}
	}

	public void initRoom(byte[] room)
	{
		int n, m;
	
		for (n = roomSize() - 1; n >= 0; n--) room[n] = 0;
		System.arraycopy(room, JSW64_PTL_PATCH, 
				no_patch, 0, no_patch.length);
		for (n = 0; n < 32; n++) room[JSW64_TITLE  + n] = 0x20;
		poke_w(room, JSW64_GPTR, 0x8000);
		room[JSW64_AIR_MAJOR] = (byte)0xFF;
		room[JSW64_AIR_MINOR] = (byte)0xFF;
		room[0] = (byte)0xFF;	// No guardians	
		m = cellTypes();
		for (n = 0; n < m; n++)
		{
			cellSetType(room, n, n);
			room[cellBase() + 9 * n] = (byte)0x80;
				/* Flashing black on black */
		}
	}

/* Convert JSW48 cell types to JSW64 */
	void jsw48CellTypes(byte [] room48, byte [] room64)
	{
		int rramp, lramp, rconv, lconv;
		int n;

		/* For engines X and V: Set up cell types that match JSW48 */
		cellSetType(room64, 0, AIR);
		cellSetType(room64, 1, WATER);
		cellSetType(room64, 2, EARTH);
		cellSetType(room64, 3, FIRE);
		cellSetType(room64, 4, LRAMP);
		cellSetType(room64, 5, LCONVEYOR);
		cellSetType(room64, 6, CRUMBLY);
		cellSetType(room64, 7, WATER);
		if ((room48[JSW48_RAMP_DIR] & 1)  != 0)  
			cellSetType(room64, 4, RRAMP);
		if ((room48[JSW48_CONVEY_DIR] & 1) != 0)
		       	cellSetType(room64, 5, RCONVEYOR);
	
	/* Copy air, earth, fire, water */
	System.arraycopy(room48, JSW48_ELEM_BASE, 
			 room64, cellLookup(room64, AIR, 1), 4 * 9);

	rramp = cellLookup(room64, RRAMP, 1);
	lramp = cellLookup(room64, LRAMP, 1);
	rconv = cellLookup(room64, RCONVEYOR, 1);
	lconv = cellLookup(room64, LCONVEYOR, 1);

	/* Migrate ramps. If the ramp goes right, then copy it to
	 * the RH ramp. If not, to the LH ramp. */
	if ((room48[JSW48_RAMP_DIR] & 1) != 0)
	{
		System.arraycopy(room48, JSW48_ELEM_RAMP, room64, rramp, 9);
		if (lramp >= 0) for (n = 1; n <= 8; n++)
			room64[lramp + n] = mirror(room64[rramp + n]);		
	}
	else
	{
		System.arraycopy(room48, JSW48_ELEM_RAMP, room64, lramp, 9);
		if (rramp >= 0) for (n = 1; n <= 8; n++)
			room64[rramp + n] = mirror(room64[lramp + n]);		
	}
	/* Migrate conveyors in the same way */
	if ((room48[JSW48_CONVEY_DIR] & 1) != 0)      /* Conveyor goes right */
	{
		System.arraycopy(room48, JSW48_ELEM_CONVEY, room64, rconv, 9);
		if (lconv >= 0) for (n = 1; n <= 8; n++) 
			room64[lconv + n] = mirror(room64[rconv + n]);
	}
	else
	{
		System.arraycopy(room48, JSW48_ELEM_CONVEY, room64, lconv, 9);
		if (rconv >= 0) for (n = 1; n <= 8; n++) 
			room64[rconv + n] = mirror(room64[lconv + n]);
	}
	/* Migrate escalators */
	if (room48[0xC4] == room48[0xCD])
	{
		if ((room48[0xD6] & 1) == (room48[0xDA] & 1))
			room64[0xFF] |= 4;
		else	room64[0xFF] |= 6;
	}
	/* Migrate the ramps in rooms like Rescue Esmerelda */
	if (room48[JSW48_ELEM_RAMP] == room48[JSW48_ELEM_WATER])
	{
		int wcell = cellLookup(room64, WATER, 1);
		if ((room48[JSW48_RAMP_DIR] & 1) != 0) 
		{
			if (rramp >= 0) 
			    System.arraycopy(room64, wcell, room64, rramp, 9);
		}
		else 
		{
			if (lramp >= 0) 
			    System.arraycopy(room64, wcell, room64, lramp, 9);
		}
	}
}


/*
private byte stepdown(byte type)
{
	if (cellTypes() == 8) switch(type)
	{
		case WATER2: case TRAMPOLINE: case WATER4: case WATER5:
		case CRUMBLY:
				type = WATER; break;
		case FIRE2:	type = FIRE; break;
		case AIR2:	type = AIR;  break;
		case EARTH2:	type = EARTH; break;
	}
	return type;
}*/


	public int cellLookup(byte [] room, int type, int which)
	{
		int idx = cellLookupIndex(room, type, which);
		if (idx < 0) return -1;

		if (idx >= cellTypes()) return -1;

		return cellBase() + 9 * idx;
	}

	// Allow this cellmap to be amended. It lives at F4FFh in bank 7. 
	private byte cellMap[] = 
	{
		0,1,2,3,4,5,6,7,
		8,9,0,1,3,2,1,1
	};


	int cellLookupIndex(byte [] room, int type, int which)
	{
		int n;

		switch (gameFormat)
		{
			case 'v':
			case 'x':
				for (n = 0; n < 4; n++)
				{
					if ((room[0x69 + n] & 0x0F) == type)
					{
						if (--which == 0)
							return 2 * n;
					}
					if (((room[0x69 + n] & 0xF0) >> 4) == type)
					{
						if (--which == 0)
							return 2 * n + 1;
					}
				}
				return -1;
			default:
				for (n = 0; n < 16; n++)
				{
					if (cellMap[n] == type)
					{
						if (--which == 0) return n;
					}
				}
				return -1;
		}
	}

	public void clearItems()
	{
		runtime[JSW64_ITEM_COUNT] = 0;
	}

	public void setStartRoom(int nr, int x, int y)
	{
		runtime[JSW64_INITIAL_ROOM] = (byte)nr;
		runtime[JSW64_INITIAL_POS1] = (byte)(y << 4);
		runtime[JSW64_INITIAL_POS2] = (byte)(((y << 5) & 0xE0) | (x & 0x1F));
		runtime[JSW64_INITIAL_POS2+1] &= 0xFE;
		runtime[JSW64_INITIAL_POS2+1] |= ((y >> 3) & 1);
	}

	public void addItem(int room, int x, int y)
	{
		int doffset;

		--runtime[JSW64_ITEM_COUNT];
		doffset = runtime[JSW64_ITEM_COUNT] & 0xFF;

		runtime[JSW64_ITEM_TABLE + doffset] = (byte)(room & 0x3F);
		if ((y & 8) != 0)
			runtime[JSW64_ITEM_TABLE + doffset] |= 0x80; /* MSB of Y */
		runtime[JSW64_ITEM_TABLE + doffset + 256] = 
			(byte)( (x & 0x1F) + ((y << 5)&0xE0));

		if (gameFormat == 'v' || gameFormat == 'w')
       		{
			runtime[JSW64_ITEM_TABLE2 + doffset] = (byte)room;
		}
	}


	public void setCell(byte [] room, int x, int y, int type, int idx)
	{
		int bits =  roomBits();
		int nbits = roomCellBits();
		byte smask, dmask;
		int n, bitsOffset;

		/* Try to find what this cell type is encoded as */
		do
		{
			n = cellLookupIndex(room, type, idx);
			if (n >= 0) break;
			--idx;
			/* If not found, cell type just isn't used in this 
			 * room. */
			if (idx == 0) return;	
		}
	 	while (n < 0);
		type = n;

		/* If 8 bits per room, we store the attribute byte for the 
		 * element */	
		if (nbits == 8)
		{
			byte cell = room[cellBase() + 9 * type];
			room[bits + y * 32 + x] = cell;
			return;
		}
		/* Else we store the element code. */
		bitsOffset = ((y * 32) + x) * nbits;
		bits += (bitsOffset / 8);
		dmask = (byte)(0x80 >> (bitsOffset & 7));
		smask = (byte)(1 << (nbits-1));

		for (n = 0; n < nbits; n++)
		{
			if ((type & smask) !=0)	room[bits] |= dmask;
			else			room[bits] &= ~dmask;
			smask = (byte)((smask & 0xFF) >> 1);
			dmask = (byte)((dmask & 0xFF) >> 1);
			if (dmask == 0)
			{
				dmask = (byte)0x80;
				++bits;
			}
		}
	}


	/** Convert JSW48 room bitmap to JSW64 */
	private void jsw48RoomBits(byte []room48, byte []room64)
	{
		int x, y, n;
		int source = 0;
		int cmap[] = new int[4];

		for (n = 0; n < 4; n++) cmap[n] = n;
		// If water = ramp, map it all as ramps
		if (room48[JSW48_ELEM_RAMP] == room48[JSW48_ELEM_WATER])
		{
			if ((room48[JSW48_RAMP_DIR] & 1) != 0)
				cmap[1] = RRAMP;
			else	cmap[1] = LRAMP;
			room64[cellLookup(room64, WATER, 1)] = (byte)0xFF;
		}
		for (y = 0; y < 16; y++)
		{
			for (x = 0; x < 32; x += 4)
			{
				setCell(room64, x,   y, cmap[3 & (room48[source] >> 6)], 1);
				setCell(room64, x+1, y, cmap[3 & (room48[source] >> 4)], 1);
				setCell(room64, x+2, y, cmap[3 & (room48[source] >> 2)], 1);
				setCell(room64, x+3, y, cmap[3 &  room48[source]], 1);
				++source;
			}
		}
	/* Convert the conveyor. */
		if (room48[JSW48_CONVEY_LEN] != 0)
		{
			x = (peek_w(room48, JSW48_CONVEY_POS) & 0x1F);
			y = (peek_w(room48, JSW48_CONVEY_POS) & 0x1FF) >> 5;
			for (n = 0; n < room48[JSW48_CONVEY_LEN]; n++,x++)
			{
				if ((room48[JSW48_CONVEY_DIR] & 1) != 0)
				{
					setCell(room64, x, y, RCONVEYOR, 1);
				}
				else
				{
					setCell(room64, x, y, LCONVEYOR, 1);
				}
			}
		}
	/* Convert the ramp. */
		if (room48[JSW48_RAMP_LEN] != 0)
		{
			x = (peek_w(room48, JSW48_RAMP_POS) & 0x1F);
			y = (peek_w(room48, JSW48_RAMP_POS) & 0x1FF) >> 5;
			for (n = 0; n < room48[JSW48_RAMP_LEN]; n++)
			{
				if ((room48[JSW48_RAMP_DIR] & 1) != 0)
				{	/* Ramp goes right */
					setCell(room64, x, y, RRAMP, 1);
					--y; 
					++x;
				}
				else
				{
					setCell(room64, x, y, LRAMP, 1);
					--y; 
					--x;
				}
			}
		}
	}

	private int jsw48Guardians(byte []room48, byte []result, int resoffset,
			byte []gtable)
	{
		int x, dcount;
		int gaddr, guard;

		gaddr = JSW48_GUARDIANS;
   		dcount = 0; 
		for (x = 0; x < 8; x++)
		{
			if (room48[gaddr] == 0xFF) break;
			guard = (8 * room48[gaddr]);
			if (result != null)
			{
				System.arraycopy(gtable, guard,
						result, resoffset+dcount, 8);
	       			result[resoffset+dcount + 2] = room48[gaddr+1];
/* Arrow. JSW64 uses bytes 1 and 3 for ink and bitmap respectively. */
				if ((gtable[guard] & 0x0F) == 4)
				{
					result[resoffset+dcount+1] = 7;	/* White arrow */
					if ((gtable[guard] & 0x80) != 0)
						result[resoffset+dcount+3] = 0x7F;
						result[resoffset+dcount+3] = (byte)0xFE;
				}	
			}
			dcount += 8;
			gaddr += 2;
		}
		if (result != null) result[resoffset + dcount] = (byte)0xFF;
		return dcount + 1;
	}


	/* Put a guardian record in the room where it'll fit best */
	public void putGuardians(byte [] room64, byte [] data)
	{
		int ps = patchSpace();
		int eb = cellBase();
		int ng;

		/* Will guardians fit before elements? */
		if (data.length <= eb)
		{
			System.arraycopy(data, 0, room64, 0, data.length);
			poke_w(room64, JSW64_GPTR, 0x8000);
			return;
		}
		/* Will they fit in patch space? */
		if (data.length <= ps)
		{
			System.arraycopy(data, 0, room64, 0x100, data.length);
			poke_w(room64, JSW64_GPTR, 0x8100);
			return;
		}
		/* Should never happen. But if it does, try to get in as many
		 * guards as possible. */
		if (eb < ps) 
		{
			ng = (ps - 1) / 8;
			System.arraycopy(data, 0, room64, 0x100, 8 * ng + 1);
			poke_w(room64, JSW64_GPTR, 0x8100);
			room64[0x100 + 8*ng] = (byte)0xFF;
		}
		else
		{
			ng = (eb - 1) / 8;
			System.arraycopy(data, 0, room64, 0, 8 * ng + 1);
			poke_w(room64, JSW64_GPTR, 0x8000);
			room64[8 * ng] = (byte)0xFF;
		}	
	}

/* Convert JSW48 guardians to JSW64 */
	private void jsw48Guardians(byte []room48, byte []room64, byte []gtable)
	{
		int required;
		byte [] ram;

		required = jsw48Guardians(room48, null, 0, gtable);
		if (required == 0) return;
		ram = new byte[required];
		jsw48Guardians(room48, ram, 0, gtable);
		putGuardians(room64, ram);
	}

	public void jsw48Room(byte [] room48, byte [] room64, byte [] gtable)
	{
		initRoom(room64);
		
		/* Copy static data */
		System.arraycopy(room48, JSW48_TITLE, 
				 room64, JSW64_TITLE, 32);
		System.arraycopy(room48, JSW48_CONVEY_DIR,
				 room64, JSW64_CONVEY_DIR, 4);
		room64[JSW64_FLAGS] = (byte)((room48[JSW48_BORDER] & 7) | 0x38);
	       					// White Willy
		System.arraycopy(room48, JSW48_ELEM_ITEM, 
				 room64, JSW64_ELEM_ITEM, 12);

		/* Copy cell types */
		jsw48CellTypes(room48, room64);
		/* Copy room bitmap */
		jsw48RoomBits(room48, room64);	
		/* Copy guardians */
		jsw48Guardians(room48, room64, gtable);
	}
 

/* Vertical guardian limit converter. Based on Andrew Broad's Java
 * conversion function (converted to C and back)  described in
 * <http://games.groups.yahoo.com/group/manicminerandjetsetwilly/message/4050>
 * and used with permission.
 */
	private void kariVguard(byte [] src, int saddr, 
				 byte [] dest, int daddr)
	{
		int velocity  = src[saddr + 4];
		byte startPos = src[saddr + 2];
		byte top      = src[saddr + 5];
		byte bottom   = src[saddr + 6];

		if (velocity < 0)
		{
			top    += (startPos - top) % velocity;
			bottom -= (bottom   - top) % velocity;
		}
		else
		{
			bottom -= (bottom - startPos) % velocity;
			top    += (bottom - top) % velocity;
		}
		if (bottom == src[saddr + 6])
		{
			bottom -= (velocity < 0) ? -velocity : velocity;
			// abs(velocity);
		}
		while (startPos > bottom)
		{
			if (velocity >= 0) velocity = -velocity;
			startPos += velocity;
		}
		if (startPos >= bottom && velocity >= 0)
		{
			velocity= -velocity;
		}
		while (startPos < top)
		{
			if (velocity < 0) velocity = -velocity;
			startPos += velocity;
		}
		if (startPos <= top && velocity < 0)
		{
			velocity = -velocity;
		}
		if (dest != null)
		{
			dest[daddr + 3] = (byte)(startPos << 1);
			dest[daddr + 4] = (byte)(velocity << 1);
			dest[daddr + 6] = (byte)(top << 1);
			dest[daddr + 7] = (byte)(bottom << 1);
		}
	}


	private void manicRoomBits(byte []room48, byte []room64)
	{
		int x, y, conveyor;
		int saddr = 0;

		if ((room48[MM_CONVEY_DIR] & 1) != 0) conveyor = RCONVEYOR;
		else			              conveyor = LCONVEYOR;

		/* Short circuit for Z-format rooms */
		if (gameFormat == 'z')
		{	
			System.arraycopy(room48, 0, room64, roomBits(), 512);
			return;
		}
		for (y = 0; y < 16; y++)
		{
			for (x = 0; x < 32; x++)
			{
				if      (room48[saddr] == room48[MM_ELEM_AIR])
					setCell(room64, x, y, AIR, 1);
				else if (room48[saddr] == room48[MM_ELEM_WATER])
					setCell(room64, x, y, WATER, 1);
				else if (room48[saddr] == room48[MM_ELEM_CRUMBLY])
					setCell(room64, x, y, CRUMBLY, 1);
				else if (room48[saddr] == room48[MM_ELEM_EARTH])
					setCell(room64, x, y, EARTH, 1);
				else if (room48[saddr] == room48[MM_ELEM_CONVEY])
					setCell(room64, x, y, conveyor, 1);
				else if (room48[saddr] == room48[MM_ELEM_FIRE])
				   	setCell(room64, x, y, FIRE, 1);
				else if (room48[saddr] == room48[MM_ELEM_FIRE2])
					setCell(room64, x, y, FIRE, 2);
				else if (room48[saddr] == room48[MM_ELEM_WATER2])
					setCell(room64, x, y, WATER, 2);
				else	setCell(room64, x, y, WATER, 3);
				++saddr;
			}
		}
	}

	/* Convert Manic Miner cell types to JSW64 */
	public void manicCellTypes(byte []room48, byte []room64)
	{
		int rconv, lconv, air, water1, earth, fire1, fire2;
		int crumbly, water2;
		int n;

		cellSetType(room64, 0, AIR);
		cellSetType(room64, 1, WATER);
		cellSetType(room64, 2, EARTH);
		cellSetType(room64, 3, FIRE);
		cellSetType(room64, 4, FIRE);
		cellSetType(room64, 5, LCONVEYOR);
		cellSetType(room64, 6, CRUMBLY);
		cellSetType(room64, 7, WATER);

		air     = cellLookup(room64, AIR, 1);
		water1  = cellLookup(room64, WATER, 1);
		earth   = cellLookup(room64, EARTH, 1);
		fire1   = cellLookup(room64, FIRE, 1);
		fire2   = cellLookup(room64, FIRE, 2);
		crumbly = cellLookup(room64, CRUMBLY, 1);
		water2  = cellLookup(room64, WATER, 2);

		System.arraycopy(room48, MM_ELEM_WATER2,  room64, water2,  9);
		System.arraycopy(room48, MM_ELEM_FIRE2,   room64, fire2,   9);
		System.arraycopy(room48, MM_ELEM_CRUMBLY, room64, crumbly, 9);
		System.arraycopy(room48, MM_ELEM_AIR,     room64, air,     9);
		System.arraycopy(room48, MM_ELEM_EARTH,   room64, earth,   9);
		System.arraycopy(room48, MM_ELEM_WATER,   room64, water1,  9);
		System.arraycopy(room48, MM_ELEM_FIRE,    room64, fire1,   9);

		/* Migrate conveyors */
		if ((room48[MM_CONVEY_DIR] & 1) != 0) /* Conveyor goes right */
		{
			cellSetType(room64, 5, RCONVEYOR);
			rconv = cellLookup(room64, RCONVEYOR, 1);
			lconv = cellLookup(room64, LCONVEYOR, 1);
			System.arraycopy(room48,MM_ELEM_CONVEY,room64,rconv,9);
			if (lconv >= 0) for (n = 1; n <= 8; n++) 
				room64[lconv + n] = mirror(room64[rconv + n]);
		}
		else
		{
			rconv = cellLookup(room64, RCONVEYOR, 1);
			lconv = cellLookup(room64, LCONVEYOR, 1);
			System.arraycopy(room48,MM_ELEM_CONVEY,room64,lconv,9);
			if (rconv >= 0) for (n = 1; n <= 8; n++) 
				room64[rconv + n] = mirror(room64[lconv + n]);
		}
		System.arraycopy(room48, MM_ELEM_ITEM, 
				 room64, JSW64_ELEM_ITEM, 8);
	}


	int manicHGuards(byte []room48, byte []guards, int guard0,
			byte spritePage, byte flags, int frames)
	{
		int n, count, base;

		count = guard0;
		base = MM_HGUARDS;
		for (n = 0; n < 5; n++)
		{
			if (room48[base] == 0)   
			{
				base += 7;
				continue;	// Blank guardian
			}
			if (room48[base] == 255) break;	/* Last guardian */
			/* The guardian whose bounds change */
			if (n == 1 && ((flags & MM_FLAG_KONG) != 0))
			{
				count += 8;
				if (guards != null)
				{
					guards[count+0] = (byte)0x88; // Trigger 
					guards[count+1] = 0x0C;
					guards[count+2] = (byte)0x80; 
						// Reference to opening wall 
					guards[count+3] = (byte)0x81; // Horizontal 
					guards[count+4] = 0;	// Speed 
					guards[count+5] = spritePage;
					guards[count+6] = (byte)(room48[base+5] & 0x1F);
								// Left limit
					guards[count+7] = 0x12; // Right limit
					if ((room48[base] & 0x80) != 0) // Speed
						guards[count+7] |= 0x20; 
					count += 8;	
				}
			}
			if (guards != null) 
			{
				guards[count] = 1;	// Horizontal
				if (room48[base+4] <= 3) 
					guards[count] |= 0x80; // Direction 
				guards[count+1] = (byte)(room48[base] & 7); // Ink
				if ((room48[base] & 0x40) != 0) 
					guards[count+1] |= 8;	// Bright 
				if (frames <= 4) guards[count+1] |= 0x60;
				else		 guards[count+1] |= 0xE0;	
				guards[count+2] = (byte)(room48[base+1] & 0x1F); // X
				if (room48[base+4] >= 4) ++guards[count+2];
				// Initial frame
				guards[count+2] |= 32 * (room48[base+4] & 8); 
				// Y
				guards[count+3] = (byte)(((room48[base+1] & 0xE0) >> 1)
				        	| ((room48[base+3] << 4)& 0x80));
				guards[count+4] = 0;	// Vertical speed
				guards[count+5] = spritePage;	// Sprite
				// Limits
				guards[count+6] = (byte)(room48[base+5] & 0x1F);
				guards[count+7] = (byte)(room48[base+6] & 0x1F);
				if ((room48[base] & 0x80) != 0)
					guards[count + 7] |= 0x20; // H Speed
			}
			base  += 7;
			count += 8;
		}	
		return count - guard0;	
	}


	int manicVGuards(byte []room48, byte []guards, int guard0,
			byte spritePage, byte flags)
	{
		int n, count, base;

		count = guard0;
		base = MM_VGUARDS;

		if ((flags & MM_FLAG_VGUARD) == 0) return 0;

		for (n = 0; n < 5; n++)
		{
			if (room48[base] == 255) break;	/* Last guardian */
			if (guards != null) 
			{
				guards[count] = 0x12;		     // Vertical
				guards[count+1] = (byte)(room48[base] & 7);  // Ink
				if ((room48[base] & 0x40) != 0) 
					guards[count+1] |= 8;	// Bright
				guards[count+1] |= 0x60;	// Frames
				guards[count+2] = (byte)(room48[base+3] & 0x1F);
			       					// X
				// Initial frame
				guards[count] |= 32 * (room48[base +1] & 3);
				guards[count+2] |= 0x80;     // Use sprites 4-7
				guards[count+5] = spritePage;  // Sprite
				kariVguard(room48, base, guards, count);
			}
			base += 7;
			count += 8;
		}	
		return count - guard0;	
	}


	int manicSkylabs(byte []room48, byte []guards, int guard0,
			byte spritePage, byte flags)
	{
		int n, count, base;

		if ((flags & MM_FLAG_SKYLAB) == 0) return 0;

		count = guard0;
		base  = MM_VGUARDS;

		for (n = 0; n < 5; n++)
		{
			if (room48[base] == 255) break;	/* Last guardian */
			if (guards != null) 
			{
// Convert to new JSW64-style skylabs
				guards[count]   = 0x08;	/* Skylab */
				guards[count+1] = (byte)(room48[base] & 7); 
			       				// Ink
				if ((room48[base] & 0x40) != 0)
					guards[count+1] |= 8;	// Bright
				guards[count+1] |= 0x80; // Move right by 8 
				guards[count+2] = (byte)(room48[base+3] & 0x1F);
			       					// X
				guards[count+3] = 0;		// Initial Y
				guards[count+4] = (byte)(room48[base+4] * 2);
			       					// dY
				guards[count+5] = spritePage;	// Sprite 
				guards[count+6] = 0;
				guards[count+7] = (byte)(2 * room48[base+6]); 
								// Y limit 
			}
			base += 7;
			count += 8;
		}
		return count - guard0;
	}		


/* Eugene becomes a 16-byte guard in JSW64. */
	int manicEugene(byte []room48, byte []guards, int guard0,
			byte spritePage, byte flags)
	{

		if ((flags & MM_FLAG_EUGENE) == 0) return 0;

		if (guards != null)
		{
			guards[guard0+0] = (byte)0x88; /* Trigger */
			guards[guard0+1] = 0x53; /* Count of untaken */
			guards[guard0+2] = (byte)0x85; /* items */
			guards[guard0+3] = 0x28; /* Angry Eugene */
			guards[guard0+4] = 0x02; /* Moving down */
			guards[guard0+5] = (byte)0xA2; /* Eugene sprite page */
			guards[guard0+6] = (byte)0xAE; /* Eugene bottom limit */
			guards[guard0+7] = (byte)0xAE; /* Eugene bottom limit */
			guards[guard0+8] = 0x02; /* Eugene as placid VG */
			guards[guard0+9] = 0x07; /* Ink */
			guards[guard0+10] = (byte)(0x0F | 0xE0); /* X | Frame*/
			guards[guard0+11] = 0x02; /* Y */
			guards[guard0+12] = 0x02; /* Moving down */
			guards[guard0+13] = (byte)0xA2; /* Eugene sprite page */
			guards[guard0+14] = 0x02; /* Eugene top limit */
			guards[guard0+15] = (byte)0xAE; /* Eugene bottom limit */
		}
		return 16;
	}


// In JSW64, Kong is rendered as 3 guardians - 2 switches and an opening
// wall. Kong himself uses patch vectors.
	int manicKong(byte []room48, byte []guards, int guard0,
			byte spritePage, byte flags)
	{
		if ((flags & MM_FLAG_KONG) == 0) return 0; /* for the 0xFF */

		if (guards != null)
		{
			guards[guard0+0] = (byte)0x98; /* Switch */
			guards[guard0+1] = 0x06; /* Ink */
			guards[guard0+2] = (byte)(0x06 | 0x80); /* X coord, and Frame = 4 */
			guards[guard0+3] = 0x00; /* Y */
			guards[guard0+4] = 0x01; /* Switch points L */
			guards[guard0+5] = (byte)0xA6; /* Eugene sprite page */
			guards[guard0+6] = 0x00; /* Switch goes right */
			guards[guard0+7] = 0x00; /* And it stays right */
			guards[guard0+8] = (byte)0xA8; /* Opening wall */
			guards[guard0+9] = 0x04; /* Triggered by previous sw. */
			guards[guard0+10] = (byte)0x80; /* ditto */
			guards[guard0+11] = 0x10; /* Height */
			guards[guard0+12] = 0x10; /* Ditto  */
			guards[guard0+13] = 0x00; /* Starts closed */
			guards[guard0+14] = 0x11; /* X-coord */
			guards[guard0+15] = (byte)0xB0; /* Y-coord */
			guards[guard0+16] = (byte)0x98; /* Switch */
			guards[guard0+17] = 0x06; /* Ink */
			guards[guard0+18] = (byte)(0x12 | 0x80); /* X coord, and Frame = 4 */
			guards[guard0+19] = 0x00; /* Y */
			guards[guard0+20] = 0x01; /* Switch points L */
			guards[guard0+21] = (byte)0xA6; /* Eugene sprite page */
			guards[guard0+22] = 0x00; /* Switch goes right */
			guards[guard0+23] = 0x00; /* And it stays right */
		}
		return 24;
	}


	public void manicRoom(byte [] room48, byte [] room64, byte spritePage,
			byte flags, int jswRoomNo, byte [] kongswitch)
	{
		int n, guard_size, guard_frames;
		byte [] guards;
		int gaddr, saddr, daddr;

		/* Pick up my patch */
		flags |= (room48[MM_BORDER] & 0xF8);
		initRoom(room64);

		System.arraycopy(room48, MM_TITLE, room64, JSW64_TITLE, 32);
		manicCellTypes(room48, room64);
		manicRoomBits(room48, room64);

		/* If in a Kong room, remove the Manic Miner switches */	
		if ((flags & MM_FLAG_KONG) != 0)
		{
			int w2 = cellLookup(room64, WATER, 2);
			if (w2 >= 0)
			{		
				setCell(room64,  6, 0, AIR, 1);
				setCell(room64, 18, 0, AIR, 1);
				for (n = 0; n < 8; n++)
				{
					kongswitch[n*2   ] = room48[w2 + n+1];
					kongswitch[n*2+ 1] = 0;
					kongswitch[n*2+16] = mirror(room48[w2 + n + 1]);
					kongswitch[n*2+17] = 0;
					room48[w2 + n + 1] = 0;
				}
				room48[w2] = (byte)0xFF;
			}	
		}

		room64[JSW64_CONVEY_DIR]   = room48[MM_CONVEY_DIR];
		room64[JSW64_CONVEY_POS]   = room48[MM_CONVEY_POS];
		room64[JSW64_CONVEY_POS+1] = (byte)(((room48[MM_CONVEY_POS+1] & 8) / 8) + 0x5E);
		room64[JSW64_CONVEY_LEN]   = room48[MM_CONVEY_LEN];

		room64[JSW64_FLAGS] = (byte)((room48[MM_BORDER] & 7) | 0x38);
		if ((flags & MM_FLAG_SOLAR) != 0)
		{
			room64[JSW64_SOLAR_X] = 23;	/* Start at X=23 Y=0 */
			room64[JSW64_SOLAR_A] = 0x77;	/* Bright White on Yellow */
		}
		if ((flags & (MM_FLAG_KONG | MM_FLAG_VGUARD)) != 0) 
			guard_frames = 3;
		else    guard_frames = 7;

		/* Items */
		for (n = 0; n < 5; n++)
		{
			saddr = MM_ITEMS + 5 * n;
			if (room48[saddr] == 0) continue;	/* No item */
			if (room48[saddr] == 0xFF) break;	/* End of table */
			--runtime[JSW64_ITEM_COUNT];
			daddr = runtime[JSW64_ITEM_COUNT] & 0xFF;
	
			runtime[JSW64_ITEM_TABLE + daddr] = (byte)(jswRoomNo & 0x3F);	/* Room number */
			if ((room48[saddr + 2] & 1) != 0)
			{	
				runtime[JSW64_ITEM_TABLE + daddr] |= 0x80; /* MSB of Y */
			}
			runtime[JSW64_ITEM_TABLE + daddr + 256] = room48[saddr + 1];      /* X and Y 
									   of item */
			if (gameFormat == 'v' || gameFormat == 'w')
			{
				runtime[JSW64_ITEM_TABLE2 + daddr] = (byte)jswRoomNo;
			}
		}
	
	
		/* Guardians */
		guard_size = manicKong   (room48, null, 0, spritePage, flags)
			   + manicHGuards(room48, null, 0, spritePage, flags, guard_frames) 
			   + manicVGuards(room48, null, 0, spritePage, flags)
			   + manicSkylabs(room48, null, 0, spritePage, flags)
			   + manicEugene (room48, null, 0, spritePage, flags)
			   + 1; /* Terminating 0xFF */
			
		guards = new byte[guard_size];
		gaddr = 0;
		gaddr += manicKong   (room48, guards, gaddr, spritePage, flags);
		gaddr += manicHGuards(room48, guards, gaddr, spritePage, flags, guard_frames);
		gaddr += manicVGuards(room48, guards, gaddr, spritePage, flags);
		gaddr += manicSkylabs(room48, guards, gaddr, spritePage, flags);
		gaddr += manicEugene (room48, guards, gaddr, spritePage, flags);
		guards[gaddr] = (byte)0xFF;	
		putGuardians(room64, guards);
		// Fixup guardians if this is a kong room
		// 0 = Left Switch
		// 1 = Opening Wall
		// 2 = Right Switch
		// 3 = Normal HG
		// 4 = HG which changeth
		if ((flags & MM_FLAG_KONG) != 0)
		{
			int gtable = peek_w(room64, JSW64_GPTR); 

			// Set the opening wall to watch the switch.
			poke_w(room64, (gtable +  9) & 0x3FF, gtable + 4);
			// Set the horizontal guardian to watch the opening 
			// wall.
			poke_w(room64, (gtable + 33) & 0x3FF, gtable + 12);
		}
		room64[JSW64_AIR_MAJOR] = room48[MM_AIR_MAJOR];
		room64[JSW64_AIR_MINOR] = room48[MM_AIR_MINOR];
		room64[JSW64_PTL_ATTR] = room48[MM_PORTAL_ATTR];
		System.arraycopy(room48, MM_PORTAL_POS, room64, JSW64_PTL_POS, 4);

		/* Exits */
		room64[JSW64_EXIT_L] = (byte)(jswRoomNo - 1);
		room64[JSW64_EXIT_R] = (byte)(jswRoomNo + 1);
		room64[JSW64_EXIT_U] = (byte)(jswRoomNo);
		room64[JSW64_EXIT_D] = (byte)(jswRoomNo);
	}

	public void cellSetType(byte [] room64, int e, int t)
	{
		int base;

		switch(gameFormat)
		{
			case 'v': case 'x':
				e = e % 8;
				base = 0x69 + (e/2);
				if ((e & 1) != 0)
				{
					room64[base] &= 0x0F;
					room64[base] |= (t << 4) & 0xF0;
				}
				else
				{
					room64[base] &= 0xF0;
					room64[base] |= (t & 0x0F);
				}
				break;
	
			// XXX Not yet done for formats W,Y,Z,[. For these formats
			// it's a global table rather than per-room.
		}
	}
	

	public void cellSetType(int room64, int e, int t)
	{
		int base;
		int rBase = roomBase(room64);

		switch(gameFormat)
		{
			case 'v': case 'x':
				e = e % 8;
				base = 0x69 + (e/2);
				if ((e & 1) != 0)
				{
					rooms[rBase + base] &= 0x0F;
					rooms[rBase + base] |= (t << 4) & 0xF0;
				}
				else
				{
					rooms[rBase + base] &= 0xF0;
					rooms[rBase + base] |= (t & 0x0F);
				}
				break;
	
			// XXX Not yet done for formats W,Y,Z,[. For these formats
			// it's a global table rather than per-room.
		}
	}

	public void setElementMap(byte [] b)
	{
		System.arraycopy(b, 0, cellMap, 0, 16);
		switch(getVariant())
		{
			case 'w':
			case 'y':
			case 'z': 
				System.arraycopy(b, 0, runtime7, 0x34FF, 16);
				break;	
			case '[': 
			case '$': 
				System.arraycopy(b, 0, runtime7, 0x3519, 16);
				break;	
		}
	}


	public void setCell(int room, int x, int y, int type, int idx)
	{
		int bits =  roomBits();
		int nbits = roomCellBits();
		byte smask, dmask;
		int n, bitsOffset;
		int rBase = roomBase(room);

		/* Try to find what this cell type is encoded as */
		do
		{
			n = cellLookupIndex(room, type, idx);
			if (n >= 0) break;
			--idx;
			/* If not found, cell type just isn't used in this 
			 * room. */
			if (idx == 0) return;	
		}
	 	while (n < 0);
		type = n;

		/* If 8 bits per room, we store the attribute byte for the 
		 * element */	
		if (nbits == 8)
		{
			byte cell = rooms[rBase + cellBase() + 9 * type];
			rooms[rBase + bits + y * 32 + x] = cell;
			return;
		}
		/* Else we store the element code. */
		bitsOffset = ((y * 32) + x) * nbits;
		bits += (bitsOffset / 8);
		dmask = (byte)(0x80 >> (bitsOffset & 7));
		smask = (byte)(1 << (nbits-1));

		for (n = 0; n < nbits; n++)
		{
			if ((type & smask) !=0)	rooms[rBase + bits] |= dmask;
			else			rooms[rBase + bits] &= ~dmask;
			smask = (byte)((smask & 0xFF) >> 1);
			dmask = (byte)((dmask & 0xFF) >> 1);
			if (dmask == 0)
			{
				dmask = (byte)0x80;
				++bits;
			}
		}
	}

	public void setCellPattern(int room, int type, int which, byte data[])
	{
		int offset = cellLookup(room, type, which);
		if (offset < 0) return;
		System.arraycopy(data, 0, rooms, offset, 9);
	}
	

	public boolean getCellPattern(int room, int type, int which, byte data[])
	{
		int offset = cellLookup(room, type, which);
		if (offset < 0) return false;
		System.arraycopy(rooms, offset, data, 0, 9);
		return true;
	}

	public void setItemBitmap(int room, byte data[])
	{
		int rBase = roomBase(room);
		System.arraycopy(data, 0, rooms, rBase + JSW64_ELEM_ITEM, 8);
	}
	

	public int cellLookup(int room, int type, int which)
	{
		int rBase = roomBase(room);
		int idx = cellLookupIndex(room, type, which);
		if (idx < 0) return -1;

		if (idx >= cellTypes()) return -1;

		return rBase + cellBase() + 9 * idx;
	}


	int cellLookupIndex(int room, int type, int which)
	{
		int n;
		int rBase = roomBase(room);

		switch (gameFormat)
		{
			case 'v':
			case 'x':
				for (n = 0; n < 4; n++)
				{
					if ((rooms[rBase + 0x69 + n] & 0x0F) == type)
					{
						if (--which == 0)
							return 2 * n;
					}
					if (((rooms[rBase + 0x69 + n] & 0xF0) >> 4) == type)
					{
						if (--which == 0)
							return 2 * n + 1;
					}
				}
				return -1;
			default:
				for (n = 0; n < 16; n++)
				{
					if (cellMap[n] == type)
					{
						if (--which == 0) return n;
					}
				}
				return -1;
		}
	}


	public void setConveyor(int room, int x, int y, int width, boolean left)
	{
		int rBase = roomBase(room);

		if (left) rooms[rBase + JSW64_CONVEY_DIR] = 0;
		else	  rooms[rBase + JSW64_CONVEY_DIR] = 1;
		rooms[rBase + JSW64_CONVEY_POS]   = (byte)((x & 0x1F) | ((y & 7) << 5));
		rooms[rBase + JSW64_CONVEY_POS+1] = (byte)(((y >> 3) & 1) | 0x5C);
		rooms[rBase + JSW64_CONVEY_LEN]   = (byte)width;
	}

	public void setWillyColour(int room, int col)
	{
		int rBase = roomBase(room);
		rooms[rBase + JSW64_FLAGS] &= ~0x38;
		rooms[rBase + JSW64_FLAGS] |= ((byte)((col & 7) << 3));
	}

	public void setWillySprite(int room, int sprite)
	{
		int rBase = roomBase(room);
		rooms[rBase + JSW64_WILLY_SPRITE] = (byte)sprite;
	}

	public void setRoomTitle(int room, String title)
	{
		int rBase = roomBase(room);
		int l = title.length();
		for (int n = 0; n < 32; n++)
		{
			if (n < l) 
				rooms[rBase + JSW64_TITLE + n] = (byte)(title.charAt(n));
			else	rooms[rBase + JSW64_TITLE + n] = 32;
		}
	}

	/* Put a guardian record in the room where it'll fit best */
	public void putGuardians(int room, byte [] data)
	{
		int rBase = roomBase(room);
		int ps = patchSpace();
		int eb = cellBase();
		int ng;

		/* Will guardians fit before elements? */
		if (data.length <= eb)
		{
			System.arraycopy(data, 0, rooms, rBase, data.length);
			poke_w(rooms, rBase + JSW64_GPTR, 0x8000);
			return;
		}
		/* Will they fit in patch space? */
		if (data.length <= ps)
		{
			System.arraycopy(data, 0, rooms, rBase + 0x100, 
					data.length);
			poke_w(rooms, rBase +  JSW64_GPTR, 0x8100);
			return;
		}
		/* Should never happen. But if it does, try to get in as many
		 * guards as possible. */
		if (eb < ps) 
		{
			ng = (ps - 1) / 8;
			System.arraycopy(data, 0, rooms, rBase + 0x100, 
					8 * ng + 1);
			poke_w(rooms, rBase + JSW64_GPTR, 0x8000);
			rooms[rBase + 0x100 + 8*ng] = (byte)0xFF;
		}
		else
		{
			ng = (eb - 1) / 8;
			System.arraycopy(data, 0, rooms, rBase, 8 * ng + 1);
			poke_w(rooms, rBase + JSW64_GPTR, 0x8000);
			rooms[rBase + 8 * ng] = (byte)0xFF;
		}	
	}

	public void storeSpritePage(int page, byte [] data)
	{
		System.arraycopy(data, 0, runtime, page * 256, 256);
// Allocate that page to sprites
		setMemoryMap(page, (byte)0x60);
	}


	public void setMemoryMap(int page, byte type)
	{
		if (page >= 0x98)
		{
			runtime[0x8640 + page - 0x98] = type;
		}
	}

	byte[] getSpritePage(int page)
	{
		byte [] data = new byte[256];
		System.arraycopy(runtime, page * 256, data, 0, 256);
		return data;
	}


	public void setPortalDest(int room, int target, int x, int y)
	{
		int rBase = roomBase(room);
		rooms[rBase + JSW64_PTL_TARGET] = (byte)target;
		rooms[rBase + JSW64_PTL_NEWXY] = (byte)((x & 0x1F) | (y << 5));
		rooms[rBase + JSW64_PTL_NEWXY+1] = (byte)(((y >> 3) & 1) | 0x5C);
		rooms[rBase + JSW64_PTL_NEWXY+2] = (byte)(16*y);
	}

	public void setPortalPos(int room, int x, int y)
	{
		int rBase = roomBase(room);
		rooms[rBase + JSW64_PTL_POS]   = (byte)((x & 0x1F) | (y << 5));
		rooms[rBase + JSW64_PTL_POS+1] = (byte)(((y >> 3) & 1) | 0x5C);
		rooms[rBase + JSW64_PTL_POS2]  = (byte)((x & 0x1F) | (y << 5));
		rooms[rBase + JSW64_PTL_POS2+1] = (byte)((y & 8) | 0x60);
	}


	public void setPortalSprite(int room, int addr)
	{
		int rBase = roomBase(room);
		rooms[rBase + JSW64_PTL_SPRITE  ] = ((byte)(addr));
		rooms[rBase + JSW64_PTL_SPRITE+1] = ((byte)(addr >> 8));
	}
	

	public void setPortalAttr(int room, byte attr)
	{
		int rBase = roomBase(room);
		rooms[rBase + JSW64_PTL_ATTR] = attr;
	}

	public void setKamikazeFlag(int room, boolean value)
	{
		int rBase = roomBase(room);
		if (value) rooms[rBase + JSW64_FLAGS2] |= JSW64_FLAG_KAMIKAZE;
		else	   rooms[rBase + JSW64_FLAGS2] &= ~JSW64_FLAG_KAMIKAZE;
	}

	public void setPortalPatch(int room, int value)
	{
		int rBase = roomBase(room);
		rooms[rBase + JSW64_PTL_PATCH    ] = (byte)value;
		rooms[rBase + JSW64_PTL_PATCH + 1] = (byte)(value >> 8);
	}

	public void setInitPatch(int room, int value)
	{
		int rBase = roomBase(room);
		rooms[rBase + JSW64_SETUP_PATCH    ] = (byte)value;
		rooms[rBase + JSW64_SETUP_PATCH + 1] = (byte)(value >> 8);
	}
	

	public void setLoopPatch(int room, int value)
	{
		int rBase = roomBase(room);
		rooms[rBase + JSW64_MAIN_PATCH    ] = (byte)value;
		rooms[rBase + JSW64_MAIN_PATCH + 1] = (byte)(value >> 8);
	}
	
	public void loadPatch(InputStream stream, int addr) throws IOException
	{
		byte data[] = new byte[16384];
		int len = stream.read(data);
		System.arraycopy(data, 0, runtime, addr, len);
	}


	public void loadPatch(String filename, int addr) throws IOException,
		FileNotFoundException
	{
		FileInputStream fis = new FileInputStream(filename);
		try
		{
			loadPatch(fis, addr);
		}
		finally
		{
			fis.close();
		}
	}

	public void setAir(int room, int value)
	{
		int rBase = roomBase(room);
		rooms[rBase + JSW64_AIR_MAJOR] = (byte)((value / 6) + 37);
		switch(value % 6)
		{
			case 0: rooms[rBase + JSW64_AIR_MINOR] = (byte)0x80;
				break;
			case 1: rooms[rBase + JSW64_AIR_MINOR] = (byte)0xC0;
				break;
			case 2: rooms[rBase + JSW64_AIR_MINOR] = (byte)0xE0;
				break;
			case 3: rooms[rBase + JSW64_AIR_MINOR] = (byte)0xF0;
				break;
			case 4: rooms[rBase + JSW64_AIR_MINOR] = (byte)0xF8;
				break;
			case 5: rooms[rBase + JSW64_AIR_MINOR] = (byte)0xFC;
				break;
		}
	}

	public void setFinalBarrier(int room, boolean top, int value)
	{
		int rBase = roomBase(room);
		if (top)
		{
			rooms[rBase + JSW64_CONVEY_DIR] &= 0xE3;
			rooms[rBase + JSW64_CONVEY_DIR] |= (value & 7) << 2;
		}
		else
		{
			rooms[rBase + JSW64_CONVEY_DIR] &= 0x1F;
			rooms[rBase + JSW64_CONVEY_DIR] |= (value & 7) << 5;
		}
	}

	public void setFinalBarrierGraphic(boolean which, byte [] value)
	{
		int screen, attr;

		if (which) { screen = 0x22FF; attr = 0x2AFF; }
		else	   { screen = 0x2BFF; attr = 0x33FF; }

		System.arraycopy(value, 0, runtime7, screen, 2048);
		System.arraycopy(value, 2048, runtime7, attr, 256);
	}
	
	public void setExit(int room, int exit, int target)
	{
		int rBase = roomBase(room);
		if (exit >= JSW64_EXIT_L  && exit <= JSW64_EXIT_D)
		{
			rooms[rBase + exit] = (byte)target;
		}
	}

	void poke7(int addr, byte value)
	{
		runtime7[addr - 0xC000] = value;
	}

	void poke(int addr, byte value)
	{
		runtime[addr] = value;
	}

	void poke(int addr, byte [] value)
	{
		System.arraycopy(value, 0, runtime, addr, value.length);
	}

	void poke(int addr, String value)
	{
		for (int n = 0; n < value.length(); n++)
		{
			runtime[addr + n] = (byte)(value.charAt(n));
		}
	}

	void setTitleScreen(byte [] title, int offset)
	{
		int len = title.length - offset;
		if (len > 0x1B00) len = 0x1B00;
		System.arraycopy(title, offset, runtime7, 0, len);
	}

	public void loadTitleTune(InputStream is) throws IOException
	{
		is.read(runtime7, 0x352A, 0xAD6);
	}


	public void loadTitleTune(String filename) throws IOException,
		FileNotFoundException
	{
		FileInputStream fis = new FileInputStream(filename);
		loadTitleTune(fis);
		fis.close();
	}

	private void patchRuntime(String hex)
		throws IOException, FileNotFoundException
	{
		HexPatcher hp = new HexPatcher(runtime);
		InputStream stream = getClass().getResourceAsStream(hex);
		if (stream == null) throw new FileNotFoundException(hex);
		hp.patch(stream);
		stream.close();
	}

	/** Given a snapshot of JSW, upgrade it to JSW64.
	 *   @param path The path to a snapshot of the original JSW48. */
	public void importJSW48(String path)
		throws IOException, FileNotFoundException
	{
// Load the JSW runtime
		FileInputStream fis = new FileInputStream(path);
		fis.skip(0x401B);
		fis.read(runtime, 0x8000, 0x8000);
		fis.close();
// Patch it with the correct hex patch
		switch(getVariant())
		{
			case 'v': patchRuntime("room_v.hex"); break;
			case 'w': patchRuntime("room_w.hex"); break;
			case 'x': patchRuntime("room_x.hex"); break;
			case 'y': patchRuntime("room_y.hex"); break;
			case 'z': patchRuntime("room_z.hex"); break;
			case '[': patchRuntime("room_yy.hex"); break;
		}

	}

}
