
package info.seasip.jsw64;

import java.io.*;

public class TapeWriter
{
	private OutputStream stream;

	public TapeWriter(OutputStream os)
	{
		stream = os;
	}

	/** Append the contents of a stream as the next entry in this 
	 * file. The stream must be in +3DOS format.
	 *
	 * @param filename The name to give the file in the TAP.
	 * @param is The stream containing the file to append.
	 * @exception IOException if an I/O error occurs. 
	 * @exception Plus3dosException if the stream does not start with
	 * a +3DOS header */
	public void writeStream(String filename, InputStream is)
		throws IOException, Plus3dosException
	{
		byte [] header = new byte[128];
		byte [] tapeHead = new byte[17];	
		int len;
		int n, rm;

		len = is.read(header);
		if (len < 128) 
			throw new Plus3dosException(filename);
		if (!Plus3dos.isPlus3dos(header)) 
			throw new Plus3dosException(filename);
		len =  (0xFF & header[11])        |
		      ((0xFF & header[12]) << 8)  |
		      ((0xFF & header[13]) << 16) |
		      ((0xFF & header[14]) << 24);

		byte data[] = new byte[len - 128];
		if (is.read(data) < len - 128)
		{
			throw(new IOException(filename + ": Unexpected EOF")); 
		}
		// Now write out the result. 

		tapeHead[0] = header[15];
		for (n = 0; n < 10; n++)
		{
			if (n < filename.length()) 
				tapeHead[n+1] = (byte)(filename.charAt(n));
			else	tapeHead[n+1] = 0x20;
		}
		for (n = 0; n < 6; n++)
		{
			tapeHead[n+11] = header[n+16];
		}
		writeBlock((byte)0,    tapeHead);
		writeBlock((byte)0xFF, data);
	}

	/** Append a byte array as the next entry in this TAP file. It will
	 * be saved as an entry of type 3 (binary data).
	 *
	 * @param filename The name to give the file in the TAP.
	 * @param src The array containing the data to write.
	 * @param address The load address in Spectrum memory. 
	 * @exception IOException if an I/O error occurs. */
	public void writeFile(String filename, byte [] src, int address)
		throws IOException
	{
		writeFile(filename, src, 0, src.length, address);
	}

	/** Append a byte array as the next entry in this TAP file. It will
	 * be saved as an entry of type 3 (binary data).
	 *
	 * @param filename The name to give the file in the TAP.
	 * @param src The array containing the data to write.
	 * @param offset The offset from which to start writing data.
	 * @param count The number of bytes to write.
	 * @param address The load address in Spectrum memory. 
	 * @exception IOException if an I/O error occurs. */
	public void writeFile(String filename, byte [] src, int offset, 
			int count, int address)
		throws IOException
	{
		writeFile(filename, src, offset, count, address, 0);
	}

	/** Append a byte array as the next entry in this TAP file. It will
	 * be saved as an entry of type 3 (binary data).
	 *
	 * @param filename The name to give the file in the TAP.
	 * @param src The array containing the data to write.
	 * @param offset The offset from which to start writing data.
	 * @param count The number of bytes to write.
	 * @param address The load address in Spectrum memory.
	 * @param exec The execute address (not used for binary data; provided
	 * only for compatibility). 
	 * @exception IOException if an I/O error occurs. */
	public void writeFile(String filename, byte [] src, int offset, 
			int count, int address, int exec)
		throws IOException
	{
		int n, rm;
		byte [] tapeHead = new byte[17];	
		int len;

		tapeHead[0] = 3;
		for (n = 0; n < 10; n++)
		{
			if (n < filename.length()) 
				tapeHead[n+1] = (byte)(filename.charAt(n));
			else	tapeHead[n+1] = 0x20;
		}
		tapeHead[11] = (byte)(count & 0xFF);
		tapeHead[12] = (byte)((count >> 8) & 0xFF);
		tapeHead[13] = (byte)(address & 0xFF);
		tapeHead[14] = (byte)((address >> 8) & 0xFF);
		tapeHead[15] = (byte)(exec & 0xFF);
		tapeHead[16] = (byte)((exec >> 8) & 0xFF);
		writeBlock((byte)0,    tapeHead);
		writeBlock((byte)0xFF, src, offset, count);
	}	

	/** Append a header or data block to the TAP file. 
	 * @param type The block type, normally 0 for header and -1 for data.
	 * @param data The contents of the block to write. 
	 * @exception IOException if an I/O error occurs. */
	public void writeBlock(byte type, byte [] data)
		throws IOException
	{
		writeBlock(type, data, 0, data.length);
	}

	/** Append a header or data block to the TAP file. 
	 * @param type The block type, normally 0 for header and -1 for data.
	 * @param data Where to get the data to write
	 * @param offset The offset in the array of the data to write
	 * @param count The number of bytes to write
	 * @exception IOException if an I/O error occurs. */
	public void writeBlock(byte type, byte [] data, int offset, int count)
		throws IOException
	{
		/* Calculate the checksum of the bytes we're writing */
		byte xsum = type;
		int n;
		for (n = 0; n < count; n++) xsum ^= data[n + offset];
		n = count + 2;

		stream.write(n & 0xFF);
		stream.write((n >> 8) & 0xFF);
		stream.write(type);
		stream.write(data, offset, count);
		stream.write(xsum);
	}
}

